# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


import os
from fnmatch import fnmatch
from glob import glob


def strip_header(lines, closing):
    # search for the header closing line, i.e. '# --\n'
    counter = 0
    found = False
    for line in lines:
        counter += 1
        if line == closing:
            found = True
            break
    if found:
        del lines[:counter]
        # If the header closing is not found, we assume it is not present.
    # add a header closing line
    lines.insert(0, closing)


def fix_python(fn, lines, header_lines):
    # check if a shebang is present
    do_shebang = lines[0].startswith("#!")
    # remove the current header
    strip_header(lines, "# --\n")
    # add a pylint line for test files:
    # if os.path.basename(fn).startswith('test_'):
    #     if not lines[1].startswith('#pylint: skip-file'):
    #         lines.insert(1, '#pylint: skip-file\n')
    # add new header (insert in reverse order)
    for hline in header_lines[::-1]:
        lines.insert(0, ("# " + hline).strip() + "\n")

    if not hline.startswith("# -*- coding: utf-8 -*-"):
        # add a source code encoding line
        lines.insert(0, "# -*- coding: utf-8 -*-\n")

    if do_shebang:
        lines.insert(0, "#!/usr/bin/env python\n")


def fix_c(fn, lines, header_lines):
    # check for an exception line
    for line in lines:
        if "no_update_headers" in line:
            return
    # remove the current header
    strip_header(lines, "//--\n")
    # add new header (insert must be in reverse order)
    for hline in header_lines[::-1]:
        lines.insert(0, ("// " + hline).strip() + "\n")


def fix_rst(fn, lines, header_lines):
    # check for an exception line
    for line in lines:
        if "no_update_headers" in line:
            return
    # remove the current header
    strip_header(lines, "    : --\n")
    # add an empty line after header if needed
    if len(lines[1].strip()) > 0:
        lines.insert(1, "\n")
    # add new header (insert must be in reverse order)
    for hline in header_lines[::-1]:
        lines.insert(0, ("    : " + hline).rstrip() + "\n")
    # add comment instruction
    lines.insert(0, "..\n")


def iter_subdirs(root):
    for dn, _, _ in os.walk(root):
        yield dn


def main():
    source_dirs = [".", "book", "notebooks", "selector"] + list(iter_subdirs("selector"))

    fixers = [
        ("*.py", fix_python),
        ("*.pxd", fix_python),
        ("*.pyx", fix_python),
        ("*.txt", fix_python),
        ("*.c", fix_c),
        ("*.cpp", fix_c),
        ("*.h", fix_c),
        ("*.rst", fix_rst),
    ]

    f = open("HEADER")
    header_lines = f.readlines()
    f.close()

    for sdir in source_dirs:
        print("Scanning:", sdir)
        for fn in glob(sdir + "/*.*"):
            if not os.path.isfile(fn):
                continue
            for pattern, fixer in fixers:
                if fnmatch(fn, pattern):
                    print(" Fixing:", fn)
                    with open(fn) as f:
                        lines = f.readlines()
                    fixer(fn, lines, header_lines)
                    with open(fn, "w") as f:
                        f.writelines(lines)
                    break


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Module for Similarity-Based Selection Methods.

This module contains the classes and functions for the similarity-based selection methods. To select
a diverse subset of molecules the similarity-based selection methods select the molecules such that
the similarity between the molecules in the subset is minimized. The similarity of a set of
molecules is calculated using an n-array similarity index. These indexes compare n molecules at a
time and return a value between 0 and 1, where 0 means that all the molecules in the set are
completely different and 1 means that the molecules are identical.

The ideas behind the similarity-based selection methods are described in the following papers:
    (esim)
    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3
    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00504-4
    (isim)
    TODO: Add paper

"""

import math
import random
from math import log
from typing import List, Optional, Union
import warnings

import numpy as np

from selector.methods.base import SelectionBase

__all__ = [
    "NSimilarity",
    "SimilarityIndex",
]


class NSimilarity(SelectionBase):
    r"""Select samples of vectors using n-ary similarity indexes between vectors.

    The algorithms in this class select a diverse subset of vectors such that the similarity
    between the vectors in the subset is minimized. The similarity of a set of vectors is
    calculated using an n-ary similarity index. These indexes compare n vectors (e.g. molecular
    fingerprints) at a time and return a value between 0 and 1, where 0 means that all the vectors
    in the set are completely different and 1 means that the vectors are identical.

    The algorithm starts by selecting a starting reference data point. Then, the next data point is
    selected such as the similarity value of the group of selected data points is minimized. The
    process is repeated until the desired number of data points is selected.

    Notes
    -----
    The ideas behind the similarity-based selection methods are described in the following papers:

    """

    def __init__(
        self,
        method: str = "isim",
        inv_order: int = 1,
        similarity_index: str = "RR",
        w_factor: str = "fraction",
        c_threshold: Union[None, str, int] = None,
        preprocess_data: bool = True,
    ):
        """Initialize class.

        Parameters
        ----------
        method: {"isim", "esim"}
            Method used for calculating the similarity indices. The methods are:
                - "isim": Instant Similarity
                    The instant similarity index calculates the average similarity of a set of
                    objects without the need to calculate the similarity of all the possible pairs.
                    This method is easier to use than the ``esim`` (extended similarity index) as it
                    does not require the use of weight factors or coincidence thresholds.
                - "esim": Extended Similarity
                    The extended similarity index calculates the similarity of a set of objects
                    without the need to calculate the similarity of all the possible pairs. This
                    method requires the use of weight factors and coincidence thresholds.
        inv_order: int
            Integer indicating the 1/inv_order power used to approximate the average of the
            similarity values elevated to 1/inv_order.
        similarity_index: str
            Key with the abbreviation of the similarity index that will be used to perform the
            selection.
            Possible values are:
                AC: Austin-Colwell
                BUB: Baroni-Urbani-Buser
                CTn: Consoni-Todschini
                Fai: Faith
                Gle: Gleason
                Ja: Jaccard
                Ja0: Jaccard 0-variant
                JT: Jaccard-Tanimoto
                RT: Rogers-Tanimoto
                RR: Russel-Rao
                SM: Sokal-Michener
                SSn: Sokal-Sneath n
        w_factor: {"fraction", "power_n"}
            Type of weight function that will be used for calculating the counters.
                'fraction' : similarity = d[k]/n
                            dissimilarity = 1 - (d[k] - n_objects % 2)/n_objects
                'power_n' : similarity = n**-(n_objects - d[k])
                            dissimilarity = n**-(d[k] - n_objects % 2)
                other values : similarity = dissimilarity = 1
            Only used for the 'esim' method. Default is 'fraction'.
        c_threshold: {None, 'dissimilar', int}
            Coincidence threshold used for calculating the similarity counters. A column of the
            elements is considered to be a coincidence among the elements if the number of elements
            that have the same value in that position is greater than the coincidence threshold.
                None : Default, c_threshold = n_objects % 2
                'dissimilar' : c_threshold = ceil(n_objects / 2)
                int : Integer number < n_objects
            Only used for the 'esim' method. Default is None.
        preprocess_data: bool
            Every data element must be between 0 and 1 for the similarity indexes to work. If
            preprocess_data is True, the data is scaled between 0 and 1 using a strategy that is
            compatible with the similarity indexes. If preprocess_data is False, the data is not
            scaled and it is assumed that the data is already between 0 and 1.

        """
        # check if the method is valid
        if method not in ["isim", "esim"]:
            raise ValueError(f'Method "{method}" is not available please select "isim" or "esim".')
        # check if the similarity index is valid
        if similarity_index not in _similarity_index_dict:
            raise ValueError(
                f'Similarity index "{similarity_index}" is not available. '
                f"See the documentation for the available similarity indexes."
            )
        # for the esim method, check if the w_factor is valid
        if method == "esim":
            # check if the w_factor and c_threshold are valid
            if w_factor != "fraction":
                if w_factor.split("_")[0] != "power" or not w_factor.split("_")[-1].isdigit():
                    print(
                        f'Invalid weight factor "{w_factor}" given. Using default value '
                        '"similarity = dissimilarity = 1".'
                    )
                    w_factor = False
            # check if the c_threshold is valid
            if c_threshold not in ["dissimilar", None]:
                if not isinstance(c_threshold, int):
                    raise ValueError(
                        f'Invalid c_threshold. It must be an integer or "dissimilar" or None. '
                        f"Given c_threshold = {c_threshold}"
                    )

        self.method = method
        self.inv_order = inv_order
        self.similarity_index = similarity_index
        self.w_factor = w_factor
        self.c_threshold = c_threshold
        self.preprocess_data = preprocess_data
        # create an instance of the SimilarityIndex class. It is used to calculate the similarity
        # index of the sets of selected objects.
        self.si = SimilarityIndex(
            method=self.method,
            inv_order=self.inv_order,
            sim_index=self.similarity_index,
            c_threshold=self.c_threshold,
            w_factor=self.w_factor,
        )

    def _scale_data(self, X: np.ndarray):
        r"""Scales the data between so it can be used with the similarity indexes.

        First each data point is normalized to be between 0 and 1.

        .. math::
            x_{ij} = \frac{x_{ij} - min(x_j)}{max(x_j) - min(x_j)}

        Then, the average of each column is calculated. Finally, each element of the final working
        array will be defined as

        .. math::
            w_{ij} = 1 - | x_{ij} - a_j |

        where :math:`x_{ij}` is the element of the normalized array,
        and :math:`a_j` is the average of the j-th
        column of the normalized array.

        Parameters
        ----------
        X: np.ndarray
            Array of features (columns) for each sample (rows).
        """
        min_value = np.min(X)
        max_value = np.max(X)
        # normalize the data to be between 0 and 1 for working with the similarity indexes
        normalized_data = (X - min_value) / (max_value - min_value)
        # calculate the average of the columns
        col_average = np.average(normalized_data, axis=0)

        # each element of the final working array will be defined as w_ij = 1 - | x_ij - a_j |
        # where x_ij is the element of the normalized array, and a_j is the average of the j-th
        # column of the normalized array.
        data = 1 - np.abs(normalized_data - col_average)
        return data

    def _get_new_index(
        self,
        X: np.ndarray,
        selected_condensed: np.ndarray,
        num_selected: int,
        select_from: np.ndarray,
    ) -> int:
        r"""Select a new diverse sample from the data.

        The function selects a new sample such that the similarity of the new set of selected
        samples is minimized.

        Parameters
        ----------
        X: np.ndarray
            Array of features (columns) for each sample (rows).
        selected_condensed: np.ndarray
            Columnwise sum of all the samples selected so far.
        num_selected: int
            Number of samples selected so far.
        select_from: np.ndarray
            Array of integers representing the indices of the samples that have not been selected
            yet.

        Returns
        -------
        selected: int
            Index of the new selected sample.

        """
        # check if the data was previously scaled
        if np.max(X) > 1 or np.min(X) < 0:
            raise ValueError(
                "The data was not scaled between 0 and 1. "
                "Use the _scale_data function to scale the data."
            )

        # Number of total vectors used to calculate th similarity. It is the number of samples
        # selected so far + 1, because the similarities are computed for the sets of samples after
        # a new selection is made.
        n_total = num_selected + 1

        # min value that is guaranteed to be higher than all the comparisons, this value should be a
        # warranty that a exist a set of samples with similarity lower than min_value. The max
        # possible similarity value for set of samples is 1.00.
        min_value = 1.01

        # placeholder index, initiating variable with a number outside the possible index values
        index = X.shape[0] + 1

        # for all indices that have not been selected
        for sample_idx in select_from:
            # column sum
            c_total = selected_condensed + X[sample_idx]

            # calculating similarity
            sim_index = self.si(c_total, nsamples=n_total)

            # if the sim of the set is less than the similarity of the previous diverse set,
            # update min_value and index
            if sim_index < min_value:
                index = sample_idx
                min_value = sim_index

        return index

    def calculate_medoid(self, X: np.ndarray, c_total=None) -> int:
        """Calculate the medoid of a set of real-valued vectors or binary objects.

        Parameters
        ----------
        X: np.array
            np.array of all the real-valued vectors or binary objects.
        c_total:
            np.array with the columnwise sums of the data, not necessary to provide.
        """
        # Check if the data is a np.ndarray of a list
        if not isinstance(X, np.ndarray):
            raise TypeError("Input data is not a np.ndarray, please input the right data type")
        # Check if the data is one dimensional
        if X.ndim != 2:
            raise ValueError(
                "Data must be a two dimensional np.ndarray for calculating the medoid."
            )
        # Check if the data has at least 3 rows
        if X.shape[0] < 3:
            raise ValueError("Input data must have at least 3 rows to calculate the medoid.")

        # check if c_total is provided and if not, calculate it
        if c_total is None:
            c_total = np.sum(X, axis=0)
        # if c_total is provided, check if it has the same number of columns as the data
        elif c_total is not None and len(X[0]) != len(c_total):
            raise ValueError("Dimensions of objects and columnwise sum differ")

        # get the total number of objects
        nsamples = X.shape[0]

        # Initialize the selected index with a number outside the possible index values
        index = nsamples + 1

        # minimum similarity value that is guaranteed to be higher than all the comparisons, this
        # value should be a warranty that a exist a sample with similarity lower than min_sim. The
        # max possible similarity value for set of samples is 1.00.
        min_sim = 1.01

        # For each sample in the set, calculate the column-wise sum of the data without the sample
        comp_sums = c_total - X

        # The medoid is calculated using an instance of the SimilarityIndex class with the same
        # parameters as the current class, but with the inv_order = 1
        si = SimilarityIndex(
            method=self.method,
            inv_order=1,
            sim_index=self.similarity_index,
            c_threshold=self.c_threshold,
            w_factor=self.w_factor,
        )

        # for each sample calculate the similarity index of the complete set without the sample
        for idx, row in enumerate(comp_sums):
            # calculate the similarity index of the set of objects without the current object
            sim_index = si(row, nsamples=nsamples - 1)
            # if the similarity is lower than the previous minimum similarity, update the minimum
            # similarity and the index
            if sim_index < min_sim:
                min_sim, index = sim_index, idx
        # the index of the object that increases more the similarity of the set when added is
        # returned (the medoid)
        return index

    def calculate_outlier(self, X: np.ndarray = None, c_total=None) -> int:
        r"""Calculate the outlier of a set of real-valued vectors or binary objects.

        Calculates the outlier of a set of real-valued vectors or binary objects. Using the
        similarity index provided in the class initialization.

        Parameters
        ----------
        X: np.array
            np.array of all the real-valued vectors or binary objects.
        c_total: np.array, optional
            np.array with the column-wise sums of the data.
        """
        # Check if the data is a np.ndarray of a list
        if not isinstance(X, np.ndarray):
            raise TypeError("Input data is not a np.ndarray, please input the right data type")
        # Check if the data is one dimensional
        if X.ndim != 2:
            raise ValueError(
                "Data must be a two dimensional np.ndarray for calculating the outlier."
            )
        # Check if the data has at least 3 rows
        if X.shape[0] < 3:
            raise ValueError("Input data must have at least 3 rows to calculate the outlier.")

        # check if c_total is provided and if not, calculate it
        if c_total is None:
            c_total = np.sum(X, axis=0)
        # if c_total is provided, check if it has the same number of columns as the data
        elif c_total is not None and len(X[0]) != len(c_total):
            raise ValueError("Dimensions of objects and columnwise sum differ")

        nsamples = X.shape[0]

        # Initialize the selected index with a number outside the possible index values
        index = nsamples + 1

        # maximum similarity value that is guaranteed to be lower than all the comparisons, this
        # value should be a warranty that a exist a sample with similarity lower than min_sim. The
        # min possible similarity value for set of samples is 0.00.
        max_sim = -0.01

        # For each sample in the set, calculate the columnwise sum of the data without the sample
        comp_sums = c_total - X

        # The outlier is calculated using an instance of the SimilarityIndex class with the same
        # parameters as the current class, but with the inv_order = 1

        # for each sample calculate the similarity index of the complete set without the sample
        for idx, obj in enumerate(comp_sums):
            # calculate the similarity index of the set of objects without the current object
            sim_index = self.si(X=obj, nsamples=nsamples - 1)
            # if the similarity is bigger than the previous minimum similarity, update the minimum
            # similarity and the index
            if sim_index > max_sim:
                max_sim, index = sim_index, idx
        # the index of the object that decreases more the similarity of the set when added is
        # returned (the outlier)
        return index

    def select_from_cluster(
        self,
        X: np.ndarray,
        size: int,
        labels: Optional[np.ndarray] = None,
        start: Union[str, List[int]] = "medoid",
    ) -> List[int]:
        r"""Algorithm of nary similarity selection for selecting points from cluster.

        Parameters
        ----------
        X: np.ndarray
            Array of features (columns) for each sample (rows).
        size: int
            Number of sample points to select (i.e. size of the subset).
        labels: np.ndarray, optional
            Array of integers or strings representing the points ids of the data that belong to the
            current cluster. If `None`, all the samples in the data are treated as one cluster.
        start: str or list
            srt: key on what is used to start the selection {'medoid', 'random', 'outlier'}.
            list: indices of points that are included in the selection since the beginning.

        Returns
        -------
        selected: list
            Indices of the selected sample points.

        """
        # check for valid start value and raise an error if it is not
        if start not in ["medoid", "random", "outlier"]:
            if not isinstance(start, list) or not all(isinstance(i, int) for i in start):
                raise ValueError(
                    "Select a correct starting point: medoid, random, outlier or a list of indices."
                )
        # check if cluster_ids are provided
        if labels is not None:
            # extract the data corresponding to the cluster_ids
            X = np.take(X, labels, axis=0)

        # total number of objects in the current cluster
        nsamples = X.shape[0]
        # ids of the data points in the current cluster (0, 1, 2, ..., samples-1)
        data_ids = np.array(range(nsamples))

        # check if the number of selected objects is less than the total number of objects
        if size > nsamples:
            raise ValueError(
                f"Number of samples is less than the requested sample size: {nsamples} < {size}."
            )

        # The data is marked to be preprocessed scale the data between 0 and 1 using a strategy
        # that is compatible with the similarity indexes
        if self.preprocess_data:
            X = self._scale_data(X)
        else:
            # check if the data is between 0 and 1 and raise an error if it is not
            if np.max(X) > 1 or np.min(X) < 0:
                raise ValueError(
                    "The data was not scaled between 0 and 1. "
                    "Use the _scale_data function to scale the data."
                )

        # select the index (of the working data) corresponding to the medoid of the data using the
        # similarity index
        if start == "medoid":
            seed = self.calculate_medoid(X)
            selected = [seed]
        # select the index (of the working data)  corresponding to a random data point
        elif start == "random":
            seed = random.randint(0, nsamples - 1)
            selected = [seed]
        # select the index (of the working data) corresponding to the outlier of the data using the
        # similarity index
        elif start == "outlier":
            seed = self.calculate_outlier(X)
            selected = [seed]
        # if a list of cluster_ids is provided, select the data_ids corresponding indices
        elif isinstance(start, list):
            if labels is not None:
                # check if all starting indices are in this cluster
                if not all(label in labels for label in start):
                    raise ValueError(
                        "Some of the provided initial indexes are not in the cluster data."
                    )
                # select the indices of the data_ids that correspond to the provided starting points
                # provided from cluster_ids
                selected = [i for i, j in enumerate(labels) if j in start]
            else:
                # check if all starting indices are in the data
                if not all(label in data_ids for label in start):
                    raise ValueError("Some of the provided initial indexes are not in the data.")
                # select the indices of the data_ids that correspond to the provided starting points
                selected = start[:]
        # Number of initial objects
        num_selected = len(selected)

        # get selected samples form the working data array
        selected_objects = np.take(X, selected, axis=0)
        # Calculate the columnwise sum of the selected samples
        selected_condensed = np.sum(selected_objects, axis=0)

        # until the desired number of objects is selected a new object is selected.
        while num_selected < size:
            # indices from which to select the new data points
            select_from = np.delete(data_ids, selected)

            # Select new index. The new object is selected such that from all possible objects the
            # similarity of the set of (selected_objects + new_object) is a minimum.
            new_index = self._get_new_index(X, selected_condensed, num_selected, select_from)

            # updating column sum vector
            selected_condensed += X[new_index]

            # updating selected indices
            selected.append(new_index)
            num_selected += 1

        return selected


class SimilarityIndex:
    r"""Calculate the n-ary similarity index of a set of vectors.

    This class provides methods for calculating the similarity index of a set of vectors represented
    as a matrix. Each vector is a row in the matrix, and each column represents a feature of the
    vector. The features in the vectors must be binary or real numbers between 0 and 1.
    """

    def __init__(
        self,
        method: str = "isim",
        inv_order: int = 1,
        sim_index: str = "RR",
        c_threshold: Union[None, str, int] = None,
        w_factor: str = "fraction",
    ):
        """Initialize the class.

        Parameters
        ----------
        method : {"isim", "esim"}, optional
            Method used for calculating the similarity indexes. The methods are:
                - "isim": Instant Similarity
                    The instant similarity index calculates the average similarity of a set of
                    objects without the need to calculate the similarity of all the possible pairs.
                    This method is easier to use than the ``esim`` (extended similarity index) as it
                    does not require the use of weight factors or coincidence thresholds.
                - "esim": Extended Similarity
                    The extended similarity index calculates the similarity of a set of objects
                    without the need to calculate the similarity of all the possible pairs. This
                    method requires the use of weight factors and coincidence thresholds.
            Default is "isim".

        inv_order : int, optional
            Integer indicating the 1/inv_order power used to approximate the average of the
            similarity values elevated to 1/inv_order. This is not used for calculating the
            medoid or the outlier. Default is 1.

        sim_index : str, optional
            The key with the abbreviation of the similarity index to be used for calculations.
            Possible values are:
                - 'AC': Austin-Colwell
                - 'BUB': Baroni-Urbani-Buser
                - 'CTn': Consoni-Todschini n
                - 'Fai': Faith
                - 'Gle': Gleason
                - 'Ja': Jaccard
                - 'Ja0': Jaccard 0-variant
                - 'JT': Jaccard-Tanimoto
                - 'RT': Rogers-Tanimoto
                - 'RR': Russel-Rao
                - 'SM': Sokal-Michener
                - 'SSn': Sokal-Sneath n
            Default is 'RR'.

        c_threshold : {None, 'dissimilar', int}, optional
            The coincidence threshold used for calculating similarity counters. A position in the
            elements is considered a coincidence (coincides among all the elements considered) if
            the number of elements that have the same value in that position is greater than the
            coincidence threshold.
                - None : Default, c_threshold = n_objects % 2
                - 'dissimilar' : c_threshold = ceil(n_objects / 2)
                - int : Integer number < n_objects
            It is only used for the 'esim' method.

        w_factor : {"fraction", "power_n"}, optional
            The type of weight function to be used.
            - 'fraction' :
                similarity = d[k] / n, dissimilarity = 1 - (d[k] - n_objects % 2) / n_objects
            - 'power_n' :
                similarity = n ** -(n_objects - d[k]), dissimilarity = n ** -(d[k] - n_objects % 2)
            - other values :
                similarity = dissimilarity = 1
            Default is 'fraction'.
                other values : similarity = dissimilarity = 1
            It is only used for the 'esim' method. Default is 'fraction'.
        """
        # check if the method is valid
        if method not in ["isim", "esim"]:
            raise ValueError(f'Method "{method}" is not available please select "isim" or "esim".')

        # check if the similarity index is valid
        if sim_index not in _similarity_index_dict:
            raise ValueError(
                f'Similarity index "{sim_index}" is not available. '
                f"See the documentation for the available similarity indexes."
            )
        # for the esim method, check if the w_factor and c_threshold are valid
        if method == "esim":
            # check if the c_threshold is valid
            if c_threshold not in ["dissimilar", None]:
                if not isinstance(c_threshold, int):
                    raise ValueError(
                        f'Invalid c_threshold. It must be an integer or "dissimilar" or None. '
                        f"Given c_threshold = {c_threshold}"
                    )
            # check if the w_factor is valid
            if w_factor != "fraction":
                if w_factor.split("_")[0] != "power" or not w_factor.split("_")[-1].isdigit():
                    print(
                        f'Invalid weight factor "{w_factor}" given. Using default value '
                        '"similarity = dissimilarity = 1".'
                    )
                    w_factor = False
        self.method = method
        self.inv_order = inv_order
        self.similarity_index = sim_index
        self.w_factor = w_factor
        self.c_threshold = c_threshold

    def _calculate_counters(self, X: np.ndarray, nsamples: Optional[int] = None) -> dict:
        """Calculate 1-similarity, 0-similarity, and dissimilarity counters.

        Arguments
        ---------
        X : np.ndarray
            Array of arrays, each sub-array contains the binary or real valued vector. The values
            must be between 0 and 1. If the number of rows ==1, the data is treated as the
            columnwise sum of the objects. If the number of rows > 1, the data is treated as the
            objects.
        nsamples: int
            Number of objects, only necessary if the columnwise sum of the objects is provided
            instead of the data (num rows== 1). If the data is provided, the number of objects is
            calculated as the length of the data.

        Returns
        -------
        counters : dict
            Dictionary with the weighted and non-weighted counters.
        """
        # Check if the data is a np.ndarray of a list
        if not isinstance(X, np.ndarray):
            raise TypeError("Argument X is not a np.ndarray, please input the right data type!")
        # Check if data is a columnwise sum or the objects
        if X.ndim == 1:
            c_total = X
            # If data is a columnwise sum, check if n_objects is provided
            if nsamples is None:
                raise ValueError("Argument X is the columnwise sum, please specify nsamples")
        else:
            c_total = np.sum(X, axis=0)
            if nsamples is not None and nsamples != len(X):
                warnings.warn(
                    f"Warning, specified number of objects {nsamples} is different from the number"
                    " of objects in data {len_arr}\n"
                    "Doing calculations with",
                    nsamples,
                    "objects.",
                )
            nsamples = len(X)

        # if method is isim, calculate the counters using the instant similarity index
        if self.method == "isim":
            # calculate number of instances with common on bits (common 1s) for each column
            a_array = c_total * (c_total - 1) / 2
            # calculate number of instances with common off bits (common 0s) for each column
            off_coincidences = nsamples - c_total
            d_array = off_coincidences * (off_coincidences - 1) / 2
            # calculate number of instances with dissimilar bits for each column
            dis_array = off_coincidences * c_total

            # calculate total a, d, b + c counters
            a = np.sum(np.power(a_array, 1.0 / self.inv_order))
            d = np.sum(np.power(d_array, 1.0 / self.inv_order))
            total_dis = np.sum(np.power(dis_array, 1 / self.inv_order))

            # calculate total similarity, and total similarity + dissimilarity
            total_sim = a + d
            p = total_sim + total_dis

            # in the isim method, the counters are not weighted, so the weighted counters are equal
            # to the non-weighted counters
            w_a = a
            w_d = d
            total_w_sim = total_sim
            total_w_dis = total_dis
            w_p = p

        elif self.method == "esim":
            # Assign c_threshold
            if self.c_threshold is None:
                tmp_c_threshold = nsamples % 2
            elif self.c_threshold == "dissimilar":
                tmp_c_threshold = math.ceil(nsamples / 2)
            elif isinstance(self.c_threshold, int):
                if self.c_threshold >= nsamples:
                    raise ValueError(
                        "c_threshold cannot be equal or greater than nsamples. \n"
                        f"c_threshold = {self.c_threshold}  nsamples = {nsamples}"
                    )
                tmp_c_threshold = self.c_threshold
            else:
                raise ValueError(
                    "c_threshold must be None, 'dissimilar' or an integer. \n"
                    f"Given c_threshold = {self.c_threshold}"
                )

            # Calculate a, d, b + c
            # Calculate the positions (columns) of common on bits (common 1s) between the objects
            a_indices = 2 * c_total - nsamples > tmp_c_threshold
            # Calculate the positions (columns) common off bits (common 0s) between the objects
            d_indices = nsamples - 2 * c_total > tmp_c_threshold
            # Calculate the positions (columns) of dissimilar bits between the objects (b + c)
            # the dissimilar bits are the bits that are not common between the objects
            dis_indices = np.abs(2 * c_total - nsamples) <= tmp_c_threshold

            # Calculate the number of columns with common on bits (common 1s) between the objects
            a = np.sum(np.power(a_indices, 1.0 / self.inv_order))
            # Calculate the number of columns with common off bits (common 0s) between the objects
            d = np.sum(np.power(d_indices, 1.0 / self.inv_order))
            # Calculate the number of columns with dissimilar bits between the objects (b + c)
            total_dis = np.sum(np.power(dis_indices, 1.0 / self.inv_order))

            # calculate the weights for each column indexed as with common on bits (common 1s)
            a_w_array = self._f_s(2 * c_total[a_indices] - nsamples, nsamples)
            # calculate the weights for each column indexed as with common off bits (common 0s)
            d_w_array = self._f_s(abs(2 * c_total[d_indices] - nsamples), nsamples)
            # calculate the weights for each column indexed as with dissimilar bits
            total_w_dis_array = self._f_d(abs(2 * c_total[dis_indices] - nsamples), nsamples)

            # calculate the total weight for each type of counter
            w_a = np.sum(np.power(a_w_array, 1.0 / self.inv_order))
            w_d = np.sum(np.power(d_w_array, 1.0 / self.inv_order))
            total_w_dis = np.sum(np.power(total_w_dis_array, 1.0 / self.inv_order))

            # calculate the counters needed to calculate the similarity indexes
            total_sim = a + d
            total_w_sim = w_a + w_d
            p = total_sim + total_dis
            w_p = total_w_sim + total_w_dis

        counters = {
            "a": a,
            "w_a": w_a,
            "d": d,
            "w_d": w_d,
            "total_sim": total_sim,
            "total_w_sim": total_w_sim,
            "total_dis": total_dis,
            "total_w_dis": total_w_dis,
            "p": p,
            "w_p": w_p,
        }
        return counters

    def _f_s(self, d, n) -> float:
        """Calculate the similarity weight factor for a given number of similar objects in a set.

        Parameters
        ----------
        d : int
            Number of similar objects.
        n : int
            Total number of objects.

        Returns
        -------
        w_s : float
            Weight factor for the similarity depending on the number of objects that are similar (d)
            in a set of (n) objects.
        """
        if self.w_factor:
            # power_n case
            if "power" in self.w_factor:
                power = int(self.w_factor.split("_")[-1])
                return power ** -(n - d).astype(float)
            # fraction case
            elif self.w_factor == "fraction":
                return d / n
            else:
                raise ValueError(
                    f"w_factor must be 'fraction' or 'power_n'. \n Given w_factor = {self.w_factor}"
                )
        # default case, the similarity counters are not weighted
        return 1

    def _f_d(self, d, n) -> float:
        """Calculate the dissimilarity weight factor for a given number of similar objects in a set.

        Parameters
        ----------
        d : int
            Number of similar objects.
        n : int
            Total number of objects.

        Returns
        -------
        w_s : float
            Weight factor for the dissimilarity depending on the number of objects that are similar
            (d) in a set of (n) objects.
        """
        if self.w_factor:
            # power_n case
            if "power" in self.w_factor:
                power = int(self.w_factor.split("_")[-1])
                return power ** -(d - n % 2).astype(float)
            # fraction case
            elif self.w_factor == "fraction":
                return 1 - (d - n % 2) / n
            else:
                raise ValueError(
                    f"w_factor must be 'fraction' or 'power_n'. \n Given w_factor = {self.w_factor}"
                )
        # default case, the dissimilarity counters are not weighted
        return 1

    def __call__(self, X: np.ndarray, nsamples: int = None) -> float:
        """Calculate the similarity index of a set of vectors.

        Parameters
        ----------
        X : np.ndarray
            Array of arrays, each sub-array contains the binary or real valued vector. The values
            must be between 0 and 1. If the number of rows ==1, the data is treated as the
            columnwise sum of the objects. If the number of rows > 1, the data is treated as the
            objects.
        nsamples: int
            Number of objects in the data. Is only necessary if the data is a columnwise sum of
            the objects. If the data is not the columnwise sum of the objects, the number of objects
            is calculated as the length of the data.
        Returns
        -------
        similarity_index: float
            Similarity index of the set of vectors.
        """
        # check if arr is a np.ndarray
        if not isinstance(X, np.ndarray):
            raise TypeError("Input data is not a np.ndarray, please input the right data type")

        # if the data is a columnwise sum of the objects check that n_objects is provided
        if X.ndim == 1:
            c_total = X
            if not nsamples:
                raise ValueError("Input data is the columnwise sum, please specify nsamples")
        # if the data is not a columnwise sum of the objects, calculate the columnwise sum and the
        # number of objects
        else:
            c_total = np.sum(X, axis=0)
            if nsamples is not None and nsamples != len(X):
                warnings.warn(
                    f"Warning, specified number of objects {nsamples} is different from the number"
                    " of objects in data {len_arr}\n"
                    "Doing calculations with",
                    nsamples,
                    "objects.",
                )
            nsamples = X.shape[0]

        # calculate the counters needed to calculate the similarity indexes
        counters = self._calculate_counters(X=c_total, nsamples=nsamples)
        # calculate the similarity index
        similarity_index = _similarity_index_dict[self.similarity_index](counters)

        return similarity_index


# Utility functions section
# -------------------------

# Functions that calculate the similarity indexes. The functions are named as the similarity
# index they calculate. The _nw suffix indicates that the similarity index is not weighted.
# More information about the similarity indexes can be found in the following paper:


def _ac_nw(counters: dict) -> float:
    """Calculate the Austin-Colwell (AC) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    ac_nw = (2 / np.pi) * np.arcsin(np.sqrt(counters["total_w_sim"] / counters["p"]))
    return ac_nw


def _bub_nw(counters: dict) -> float:
    """Calculate the Baroni-Urbani-Buser (BUB) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    bub_nw = ((counters["w_a"] * counters["w_d"]) ** 0.5 + counters["w_a"]) / (
        (counters["a"] * counters["d"]) ** 0.5 + counters["a"] + counters["total_dis"]
    )
    return bub_nw


def _ct1_nw(counters: dict) -> float:
    """Calculate the Consoni-Todschini 1 (CT1) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    ct1_nw = (log(1 + counters["w_a"] + counters["w_d"])) / (log(1 + counters["p"]))
    return ct1_nw


def _ct2_nw(counters: dict) -> float:
    """Calculate the Consoni-Todschini 2 (CT2) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    ct2_nw = (log(1 + counters["w_p"]) - log(1 + counters["total_w_dis"])) / (
        log(1 + counters["p"])
    )
    return ct2_nw


def _ct3_nw(counters: dict) -> float:
    """Calculate the Consoni-Todschini 3 (CT3) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    ct3_nw = (log(1 + counters["w_a"])) / (log(1 + counters["p"]))
    return ct3_nw


def _ct4_nw(counters: dict) -> float:
    """Calculate the Consoni-Todschini 4 (CT4) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    ct4_nw = (log(1 + counters["w_a"])) / (log(1 + counters["a"] + counters["total_dis"]))
    return ct4_nw


def _fai_nw(counters: dict) -> float:
    """Calculate the Faith (Fai) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    fai_nw = (counters["w_a"] + 0.5 * counters["w_d"]) / (counters["p"])
    return fai_nw


def _gle_nw(counters: dict) -> float:
    """Calculate the Gleason (Gle) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    gle_nw = (2 * counters["w_a"]) / (2 * counters["a"] + counters["total_dis"])
    return gle_nw


def _ja_nw(counters: dict) -> float:
    """Calculate the Jaccard (Ja) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    ja_nw = (3 * counters["w_a"]) / (3 * counters["a"] + counters["total_dis"])
    return ja_nw


def _ja0_nw(counters: dict) -> float:
    """Calculate the Jaccard 0-variant (Ja0) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    ja0_nw = (3 * counters["total_w_sim"]) / (3 * counters["total_sim"] + counters["total_dis"])
    return ja0_nw


def _jt_nw(counters: dict) -> float:
    """Calculate the Jaccard-Tanimoto (JT) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    jt_nw = (counters["w_a"]) / (counters["a"] + counters["total_dis"])
    return jt_nw


def _rt_nw(counters: dict) -> float:
    """Calculate the Rogers-Tanimoto (RT) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    rt_nw = (counters["total_w_sim"]) / (counters["p"] + counters["total_dis"])
    return rt_nw


def _rr_nw(counters: dict) -> float:
    """Calculate the Russel-Rao (RR) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    rr_nw = (counters["w_a"]) / (counters["p"])
    return rr_nw


def _sm_nw(counters: dict) -> float:
    """Calculate the Sokal-Michener (SM) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    sm_nw = (counters["total_w_sim"]) / (counters["p"])
    return sm_nw


def _ss1_nw(counters: dict) -> float:
    """Calculate the Sokal-Sneath 1 (SS1) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    ss1_nw = (counters["w_a"]) / (counters["a"] + 2 * counters["total_dis"])
    return ss1_nw


def _ss2_nw(counters: dict) -> float:
    """Calculate the Sokal-Sneath 2 (SS2) similarity index.

    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00505-3"""
    ss2_nw = (2 * counters["total_w_sim"]) / (counters["p"] + counters["total_sim"])
    return ss2_nw


# Dictionary with the similarity indexes functions as values and the keys are the abbreviations
_similarity_index_dict = {
    "AC": _ac_nw,
    "BUB": _bub_nw,
    "CT1": _ct1_nw,
    "CT2": _ct2_nw,
    "CT3": _ct3_nw,
    "CT4": _ct4_nw,
    "Fai": _fai_nw,
    "Gle": _gle_nw,
    "Ja": _ja_nw,
    "Ja0": _ja0_nw,
    "JT": _jt_nw,
    "RT": _rt_nw,
    "RR": _rr_nw,
    "SM": _sm_nw,
    "SS1": _ss1_nw,
    "SS2": _ss2_nw,
}



# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --

"""Molecule dataset diversity calculation module."""

import warnings

import numpy as np

from selector.measures.similarity import tanimoto

__all__ = [
    "compute_diversity",
    "logdet",
    "shannon_entropy",
    "explicit_diversity_index",
    "wdud",
    "hypersphere_overlap_of_subset",
    "gini_coefficient",
    "nearest_average_tanimoto",
]


def compute_diversity(
    feature_subset: np.array,
    div_type: str = "shannon_entropy",
    normalize: bool = False,
    truncation: bool = False,
    features: np.array = None,
    cs: int = None,
) -> float:
    """Compute diversity metrics.

    Parameters
    ----------
    feature_subset : np.ndarray
        Feature matrix.
    div_type : str, optional
        Method of calculation diversity for a given molecule set, which
        includes "entropy", "logdet", "shannon entropy", "wdud",
        "gini coefficient" "hypersphere_overlap", and
        "explicit diversity index".
        The default is "entropy".
    normalize : bool, optional
        Normalize the entropy to [0, 1]. Default is "False".
    truncation : bool, optional
        Use the truncated Shannon entropy. Default is "False".
    features : np.ndarray, optional
        Feature matrix of entire molecule library, used only if
        calculating `hypersphere_overlap_of_subset`. Default is "None".
    cs : int, optional
        Number of common substructures in molecular compound dataset.
        Used only if calculating `explicit_diversity_index`. Default is "None".


    Returns
    -------
    float, computed diversity.

    """
    func_dict = {
        "logdet": logdet,
        "wdud": wdud,
        "gini_coefficient": gini_coefficient,
    }

    if div_type in func_dict:
        return func_dict[div_type](feature_subset)

    # hypersphere overlap of subset
    elif div_type == "hypersphere_overlap":
        if features is None:
            raise ValueError(
                "Please input a feature matrix of the entire "
                "dataset when calculating hypersphere overlap."
            )
        return hypersphere_overlap_of_subset(features, feature_subset)

    elif div_type == "shannon_entropy":
        return shannon_entropy(feature_subset, normalize=normalize, truncation=truncation)

    elif div_type == "explicit_diversity_index":
        if cs is None:
            raise ValueError(
                "Attribute `cs` is missing. "
                "Please input `cs` value to use explicit_diversity_index."
            )
        elif cs == 0:
            raise ValueError("Divide by zero error: Attribute `cs` cannot be 0.")
        return explicit_diversity_index(feature_subset, cs)
    else:
        raise ValueError(f"Diversity type {div_type} not supported.")


def logdet(x: np.ndarray) -> float:
    r"""Compute the log determinant function.

    Given a  :math:`n_s \times n_f` feature matrix :math:`x`, where :math:`n_s` is the number of
    samples and :math:`n_f` is the number of features, the log determinant function is defined as:

    .. math:
        F_\text{logdet} = \log{\left(\det{\left(\mathbf{x}\mathbf{x}^T + \mathbf{I}\right)}\right)}

    where the :math:`I` is the :math:`n_s \times n_s` identity matrix.
    Higher values of :math:`F_\text{logdet}` mean more diversity.

    Parameters
    ----------
    x: ndarray of shape (n_samples, n_features)
        Feature matrix of `n_samples` samples in `n_features` dimensional feature space,

    Returns
    -------
    f_logdet: float
        The volume of parallelotope spanned by the matrix.

    Notes
    -----
    The log-determinant function is based on the formula in Nakamura, T., Sci Rep 2022.
    Please note that we used the
    natural logrithim to avoid the numerical stability issues,
    https://github.com/theochem/Selector/issues/229.

    References
    ----------
    Nakamura, T., Sakaue, S., Fujii, K., Harabuchi, Y., Maeda, S., and Iwata, S.., Selecting
    molecules with diverse structures and properties by maximizing submodular functions of
    descriptors learned with graph neural networks. Scientific Reports 12, 2022.

    """
    mid = np.dot(x, np.transpose(x)) + np.identity(x.shape[0])
    logdet_mid = np.linalg.slogdet(mid)
    f_logdet = logdet_mid.sign * logdet_mid.logabsdet
    return f_logdet


def shannon_entropy(x: np.ndarray, normalize=True, truncation=False) -> float:
    r"""Compute the shannon entropy of a binary matrix.

    Higher values mean more diversity.

    Parameters
    ----------
    x : ndarray
        Bit-string matrix.
    normalize : bool, optional
        Normalize the entropy to [0, 1]. Default=True.
    truncation : bool, optional
        Use the truncated Shannon entropy by only counting the contributions of one-bits.
        Default=False.

    Returns
    -------
    float :
        The shannon entropy of the matrix.

    Notes
    -----
    Suppose we have :math:`m` compounds and each compound has :math:`n` bits binary fingerprints.
    The binary matrix (feature matrix) is :math:`\mathbf{x} \in m \times n`, where each
    row is a compound and each column contains the :math:`n`-bit binary fingerprint.
    The equation for Shannon entropy is given by [1]_ and [3]_,

    .. math::
        H = \sum_i^m \left[ - p_i \log_2{p_i }  - (1 - p_i)\log_2(1 - p_i) \right]

    where :math:`p_i` represents the relative frequency of `1` bits at the fingerprint position
    :math:`i`. When :math:`p_i = 0` or :math:`p_i = 1`, the :math:`SE_i` is zero.
    When `completeness` is True, the entropy is calculated as in [2]_ instead

    .. math::
        H = \sum_i^m \left[ - p_i \log_2{p_i } \right]

    When `normalize` is True, the entropy is normalized by a scaling factor so that the entropy is in the range of
    [0, 1], [2]_

    .. math::
        H = \frac{ \sum_i^m \left[ - p_i \log_2{p_i }  - (1 - p_i)\log_2(1 - p_i) \right]}
            {n \log_2{2} / 2}

    But please note, when `completeness` is False and `normalize` is True, the formula has not been
    used in any literature. It is just a simple normalization of the entropy and the user can use it at their own risk.

    References
    ----------
    .. [1] Wang, Y., Geppert, H., & Bajorath, J. (2009). Shannon entropy-based fingerprint similarity
       search strategy. Journal of Chemical Information and Modeling, 49(7), 1687-1691.
    .. [2] Leguy, J., Glavatskikh, M., Cauchy, T., & Da Mota, B. (2021). Scalable estimator of the
       diversity for de novo molecular generation resulting in a more robust QM dataset (OD9) and a
       more efficient molecular optimization. Journal of Cheminformatics, 13(1), 1-17.
    .. [3] Weidlich, I. E., & Filippov, I. V. (2016). Using the Gini coefficient to measure the
       chemical diversity of small molecule libraries. Journal of Computational Chemistry, 37(22), 2091-2097.

    """
    # check if matrix is binary
    if np.count_nonzero((x != 0) & (x != 1)) != 0:
        raise ValueError("Attribute `x` should have binary values.")

    p_i_arr = np.sum(x, axis=0) / x.shape[0]
    h_x = 0

    for p_i in p_i_arr:
        if p_i == 0 or p_i == 1:
            # p_i = 0
            se_i = 0
        else:
            if truncation:
                # from https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00554-8
                se_i = -p_i * np.log2(p_i)
            else:
                # from https://pubs.acs.org/doi/10.1021/ci900159f
                se_i = -p_i * np.log2(p_i) - (1 - p_i) * np.log2(1 - p_i)

        h_x += se_i

    if normalize:
        if truncation:
            warnings.warn(
                "Computing the normalized Shannon entropy only counting the on-bits has not been reported in "
                "literature. The user can use it at their own risk."
            )

        h_x /= x.shape[1] * np.log2(2) / 2

    return h_x


# todo: add tests for edi
def explicit_diversity_index(
    x: np.ndarray,
    cs: int,
) -> float:
    """Compute the explicit diversity index.

    Parameters
    ----------
    x: ndarray of shape (n_samples, n_features)
        Feature matrix of `n_samples` samples in `n_features` dimensional feature space.
    cs : int
        Number of common substructures in the compound set.

    Returns
    -------
    edi_scaled : float
        Explicit diversity index.

    Notes
    -----
    This method hasn't been tested.

    This method is used only for datasets of molecular compounds.

    Papp, Á., Gulyás-Forró, A., Gulyás, Z., Dormán, G., Ürge, L.,
    and Darvas, F.. (2006) Explicit Diversity Index (EDI):
    A Novel Measure for Assessing the Diversity of Compound Databases.
    Journal of Chemical Information and Modeling 46, 1898-1904.
    """
    nc = len(x)
    sdi = (1 - nearest_average_tanimoto(x)) / (0.8047 - (0.065 * (np.log(nc))))
    cr = -1 * np.log10(nc / (cs**2))
    edi = (sdi + cr) * 0.7071067811865476
    edi_scaled = ((np.tanh(edi / 3) + 1) / 2) * 100
    return edi_scaled


def wdud(x: np.ndarray) -> float:
    r"""Compute the Wasserstein Distance to Uniform Distribution(WDUD).

    The equation for the Wasserstein Distance for a single feature to uniform distribution is

    .. math::
        WDUD(x) = \int_{0}^{1} |U(x) - V(x)|dx

    where the feature is normalized to [0, 1], :math:`U(x)=x` is the cumulative distribution
    of the uniform distribution on [0, 1], and :math:`V(x) = \sum_{y <= x}1 / N` is the discrete
    distribution of the values of the feature in :math:`x`, where :math:`y` is the ith feature. This
    integral is calculated iteratively between :math:`y_i` and :math:`y_{i+1}`, using trapezoidal method.

    Lower values of the WDUD mean more diversity because the features of the selected set are
    more evenly distributed over the range of feature values.

    Parameters
    ----------
    x: ndarray of shape (n_samples, n_features)
        Feature matrix of `n_samples` samples in `n_features` dimensional feature space.

    Returns
    -------
    float :
        The mean of the WDUD of each feature over all molecules.

    Notes
    -----
    Nakamura, T., Sakaue, S., Fujii, K., Harabuchi, Y., Maeda, S., and Iwata, S.. (2022)
    Selecting molecules with diverse structures and properties by maximizing
    submodular functions of descriptors learned with graph neural networks.
    Scientific Reports 12.

    """
    if x.ndim != 2:
        raise ValueError(f"The number of dimensions {x.ndim} should be two.")

    # find the range of each feature
    col_diff = np.ptp(x, axis=0)
    # Normalization of each feature to [0, 1]
    if np.any(np.abs(col_diff) < 1e-30):
        # warning if some feature columns are constant
        warnings.warn(
            "Some of the features are constant which will cause the normalization to fail. "
            "Now removing them."
        )
        if np.all(col_diff < 1.0e-30):
            raise ValueError(
                "Unfortunately, all the features are constants and wdud cannot be calculated."
            )
        else:
            # remove the constant feature columns
            mask = np.ptp(x, axis=0) > 1e-30
            x = x[:, mask]
    x_norm = (x - np.min(x, axis=0)) / np.ptp(x, axis=0)

    # min_max normalization:
    n_samples, n_features = x_norm.shape
    ans = []  # store the Wasserstein distance for each feature
    for i in range(0, n_features):
        wdu = 0.0
        y = np.sort(x_norm[:, i])
        # Round to the sixth decimal place and count number of unique elements
        #    to construct an accurate cumulative discrete distribution func \sum_{x <= y_{i + 1}} 1/k
        y, counts = np.unique(np.round(x_norm[:, i], decimals=6), return_counts=True)
        p = 0
        # Ignore 0 and because v_min= 0
        for j in range(1, len(counts)):
            # integral from y_{i - 1} to y_{i} of |x - \sum_{x <= y_{i}} 1/k| dx
            yi1 = y[j - 1]
            yi = y[j]
            # Make a grid from yi1 to yi
            grid = np.linspace(yi1, yi, num=1000, endpoint=True)
            # Evaluate the integrand  |x - \sum_{x <= y_{i + 1}} 1/k|
            p += counts[j - 1]
            integrand = np.abs(grid - p / n_samples)
            # Integrate using np.trapz
            wdu += np.trapz(y=integrand, x=grid)
        ans.append(wdu)
    return np.average(ans)


def hypersphere_overlap_of_subset(x: np.ndarray, x_subset: np.array) -> float:
    r"""Compute the overlap of subset with hyper-spheres around each point

    The edge penalty is also included, which disregards areas
    outside of the boundary of the full feature space/library.
    This is calculated as:

    .. math::
        g(S) = \sum_{i < j}^k O(i, j) + \sum^k_m E(m)

    where :math:`i, j` is over the subset of molecules,
    :math:`O(i, j)` is the approximate overlap between hyperspheres,
    :math:`k` is the number of features and :math:`E`
    is the edge penalty of a molecule.

    Lower values mean more diversity.

    Parameters
    ----------
    x : ndarray
        Feature matrix of all molecules.
    x_subset : ndarray
        Feature matrix of selected subset of molecules.

    Returns
    -------
    float :
        The approximate overlapping volume of hyperspheres
        drawn around the selected points/molecules.

    Notes
    -----
    The hypersphere overlap volume is calculated using an approximation formula from Agrafiotis (1997).

    Agrafiotis, D. K.. (1997) Stochastic Algorithms for Maximizing Molecular Diversity.
    Journal of Chemical Information and Computer Sciences 37, 841-851.
    """

    # Find the maximum and minimum over each feature across all molecules.
    max_x = np.max(x, axis=0)
    min_x = np.min(x, axis=0)

    if np.all(np.abs(max_x - min_x) < 1e-30):
        raise ValueError("All of the features are redundant which causes normalization to fail.")

    # Remove redundant features
    non_red_feat = np.abs(max_x - min_x) > 1e-30
    x = x[:, non_red_feat]
    x_subset = x_subset[:, non_red_feat]
    max_x = max_x[non_red_feat]
    min_x = min_x[non_red_feat]

    d = len(x_subset[0])
    k = len(x_subset[:, 0])

    # normalization of each feature to [0, 1]
    x_norm = (x_subset - min_x) / (max_x - min_x)

    # r_o = hypersphere radius
    r_o = d * np.sqrt(1 / k)
    if r_o > 0.5:
        warnings.warn(
            "The number of molecules should be much larger" " than the number of features."
        )
    g_s = 0
    edge = 0

    # lambda parameter controls edge penalty
    lam = (d - 1.0) / d
    # calculate overlap volume
    for i in range(0, (k - 1)):
        for j in range((i + 1), k):
            dist = np.linalg.norm(x_norm[i] - x_norm[j])
            # Overlap penalty
            if dist <= (2 * r_o):
                with np.errstate(divide="ignore"):
                    # min(100) ignores the inf case with divide by zero
                    g_s += min(100, 2 * (r_o / dist) - 1)
        # Edge penalty: lambda (1 - \sum^d_j e_{ij} / (dr_0)
        edge_pen = 0.0
        for j_dim in range(0, d):
            # calculate dist to closest boundary in jth coordinate,
            # with max value = 1, min value = 0
            dist_max = np.abs(1 - x_norm[i, j_dim])
            dist_min = x_norm[i, j_dim]
            dist = min(dist_min, dist_max)
            # truncate distance at r_o
            if dist > r_o:
                dist = r_o
            edge_pen += dist
        edge_pen /= d * r_o
        edge_pen = lam * (1.0 - edge_pen)
        edge += edge_pen
    g_s += edge
    return g_s


def gini_coefficient(x: np.ndarray):
    r"""
    Gini coefficient of bit-wise fingerprints of a database of molecules.

    Measures the chemical diversity of a database of molecules defined by
    the following formula:

    .. math::
        G = \frac{2 \sum_{i=1}^L i ||y_i||_1 }{N \sum_{i=1}^L ||y_i||_1} - \frac{L+1}{L},

    where :math:`y_i \in \{0, 1\}^N` is a vector of zero and ones of length the
    number of molecules :math:`N` of the `i`th feature, and :math:`L` is the feature length.

    Lower values mean more diversity.

    Parameters
    ----------
    x : ndarray(N, L)
        Molecule features in L bits with N molecules.

    Returns
    -------
    float :
        Gini coefficient in the range [0,1].

    References
    ----------
    Weidlich, Iwona E., and Igor V. Filippov. "Using the gini coefficient to measure the
    chemical diversity of small‐molecule libraries." (2016): 2091-2097.

    """
    # Check that `x` is a bit-wise fingerprint.
    if np.count_nonzero((x != 0) & (x != 1)) != 0:
        raise ValueError("Attribute `x` should have binary values.")
    if x.ndim != 2:
        raise ValueError(f"Attribute `x` should have dimension two rather than {x.ndim}.")

    num_features = x.shape[1]
    # Take the bit-count of each column/molecule.
    bit_count = np.sum(x, axis=0)

    # Sort the bit-count since Gini coefficients relies on cumulative distribution.
    bit_count = np.sort(bit_count)

    # Mean of denominator
    denominator = num_features * np.sum(bit_count)
    numerator = np.sum(np.arange(1, num_features + 1) * bit_count)

    return 2.0 * numerator / denominator - (num_features + 1) / num_features


def nearest_average_tanimoto(x: np.ndarray) -> float:
    """Computes the average tanimoto for nearest molecules.

    Parameters
    ----------
    x : ndarray
        Feature matrix.

    Returns
    -------
    nat : float
        Average tanimoto of closest pairs.

    Notes
    -----
    This computes the tanimoto coefficient of pairs with the shortest
    distances, then returns the average of them.
    This calculation is explictly for the explicit diversity index.

    Papp, Á., Gulyás-Forró, A., Gulyás, Z., Dormán, G., Ürge, L.,
    and Darvas, F.. (2006) Explicit Diversity Index (EDI):
    A Novel Measure for Assessing the Diversity of Compound Databases.
    Journal of Chemical Information and Modeling 46, 1898-1904.
    """
    tani = []
    for idx, _ in enumerate(x):
        # arbitrary distance for comparison:
        short = 100
        a = 0
        b = 0
        # search for shortest distance point from idx
        for jdx, _ in enumerate(x):
            dist = np.linalg.norm(x[idx] - x[jdx])
            if dist < short and idx != jdx:
                short = dist
                a = idx
                b = jdx
        # calculate tanimoto for each shortest dist pair
        tani.append(tanimoto(x[a], x[b]))
    # compute average of all shortest tanimoto coeffs
    nat = np.average(tani)
    return nat



# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --

"""Test Diversity Module."""
import warnings

import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_equal, assert_raises, assert_warns

from selector.measures.diversity import (
    compute_diversity,
    explicit_diversity_index,
    gini_coefficient,
    hypersphere_overlap_of_subset,
    logdet,
    nearest_average_tanimoto,
    shannon_entropy,
    wdud,
)

# each row is a feature and each column is a molecule
sample1 = np.array([[4, 2, 6], [4, 9, 6], [2, 5, 0], [2, 0, 9], [5, 3, 0]])

# each row is a molecule and each column is a feature (scipy)
sample2 = np.array([[1, 1, 0, 0, 0], [0, 1, 1, 0, 0], [0, 0, 0, 1, 0], [0, 0, 0, 0, 1]])

sample3 = np.array([[1, 4], [3, 2]])

sample4 = np.array([[1, 0, 1], [0, 1, 1]])

sample5 = np.array([[0, 2, 4, 0], [1, 2, 4, 0], [2, 2, 4, 0]])

sample6 = np.array([[1, 0, 1, 0], [0, 1, 1, 0], [1, 0, 1, 0], [0, 0, 1, 0]])

sample7 = np.array([[1, 0, 1, 0] for _ in range(4)])

sample8 = np.array([[1, 2, 3], [1, 2, 3], [1, 2, 3]])


def test_compute_diversity_specified():
    """Test compute diversity with a specified div_type."""
    comp_div = compute_diversity(sample6, "shannon_entropy", normalize=False, truncation=False)
    expected = 1.81
    assert round(comp_div, 2) == expected


def test_compute_diversity_hyperspheres():
    """Test compute diversity with two arguments for hypersphere_overlap method"""
    corner_pts = np.array([[0.0, 0.0], [0.0, 1.0], [1.0, 0.0], [1.0, 1.0]])
    centers_pts = np.array([[0.5, 0.5]] * (100 - 4))
    pts = np.vstack((corner_pts, centers_pts))

    comp_div = compute_diversity(pts, div_type="hypersphere_overlap", features=pts)
    # Expected = overlap + edge penalty
    expected = (100.0 * 96 * 95 * 0.5) + 2.0
    assert_almost_equal(comp_div, expected)


def test_compute_diversity_hypersphere_error():
    """Test compute diversity with hypersphere metric and no molecule library given."""
    assert_raises(ValueError, compute_diversity, sample5, "hypersphere_overlap")


def test_compute_diversity_edi():
    """Test compute diversity with explicit diversity index div_type"""
    z = np.array([[0, 1, 2], [1, 2, 0], [2, 0, 1]])
    cs = 1
    expected = 56.39551204
    actual = compute_diversity(z, "explicit_diversity_index", cs=cs)
    assert_almost_equal(expected, actual)


def test_compute_diversity_edi_no_cs_error():
    """Test compute diversity with explicit diversity index and no `cs` value given."""
    assert_raises(ValueError, compute_diversity, sample5, "explicit_diversity_index")


def test_compute_diversity_edi_zero_error():
    """Test compute diversity with explicit diversity index and `cs` = 0."""
    assert_raises(ValueError, compute_diversity, sample5, "explicit diversity index", cs=0)


def test_compute_diversity_invalid():
    """Test compute diversity with a non-supported div_type."""
    assert_raises(ValueError, compute_diversity, sample1, "diversity_type")


def test_logdet():
    """Test the log determinant function with predefined subset matrix."""
    sel = logdet(sample3)
    expected = np.log(131)
    assert_almost_equal(sel, expected)


def test_logdet_non_square_matrix():
    """Test the log determinant function with a rectangular matrix."""
    sel = logdet(sample4)
    expected = np.log(8)
    assert_almost_equal(sel, expected)


def test_shannon_entropy():
    """Test the shannon entropy function with example from the original paper."""

    # example taken from figure 1 of 10.1021/ci900159f
    x1 = np.array([[1, 0, 1, 0], [0, 1, 1, 0], [1, 0, 1, 0], [0, 0, 1, 0]])
    expected = 1.81
    assert round(shannon_entropy(x1, normalize=False, truncation=False), 2) == expected

    x2 = np.vstack((x1, [1, 1, 1, 0]))
    expected = 1.94
    assert round(shannon_entropy(x2, normalize=False, truncation=False), 2) == expected

    x3 = np.vstack((x1, [0, 1, 0, 1]))
    expected = 3.39
    assert round(shannon_entropy(x3, normalize=False, truncation=False), 2) == expected


def test_shannon_entropy_normalize():
    """Test the shannon entropy function with normalization."""
    x1 = np.array([[1, 0, 1, 0], [0, 1, 1, 0], [1, 0, 1, 0], [0, 0, 1, 0]])
    expected = 1.81 / (x1.shape[1] * np.log2(2) / 2)
    assert_almost_equal(
        actual=shannon_entropy(x1, normalize=True, truncation=False),
        desired=expected,
        decimal=2,
    )


def test_shannon_entropy_warning():
    """Test the shannon entropy function gives warning when normalization is True and truncation is True."""
    x1 = np.array([[1, 0, 1, 0], [0, 1, 1, 0], [1, 0, 1, 0], [0, 0, 1, 0]])
    with pytest.warns(UserWarning):
        shannon_entropy(x1, normalize=True, truncation=True)


def test_shannon_entropy_binary_error():
    """Test the shannon entropy function raises error with a non binary matrix."""
    assert_raises(ValueError, shannon_entropy, sample5)


def test_explicit_diversity_index():
    """Test the explicit diversity index function."""
    z = np.array([[0, 1, 2], [1, 2, 0], [2, 0, 1]])
    cs = 1
    nc = 3
    sdi = 0.75 / 0.7332902012
    cr = -1 * 0.4771212547
    edi = 0.5456661753 * 0.7071067811865476
    edi_scaled = 56.395512045413
    value = explicit_diversity_index(z, cs)
    assert_almost_equal(value, edi_scaled, decimal=8)


def test_wdud_uniform():
    """Test wdud when a feature has uniform distribution."""
    uni = np.arange(0, 50000)[:, None]
    wdud_val = wdud(uni)
    expected = 0
    assert_almost_equal(wdud_val, expected, decimal=4)


def test_wdud_repeat_yi():
    """Test wdud when a feature has multiple identical values."""
    dist = np.array([[0, 0.5, 0.5, 0.75, 1]]).T
    wdud_val = wdud(dist)
    # calculated using wolfram alpha:
    expected = 0.065 + 0.01625 + 0.02125
    assert_almost_equal(wdud_val, expected, decimal=4)


def test_wdud_mult_features():
    """Test wdud when there are multiple features per molecule."""
    dist = np.array(
        [
            [0, 0.5, 0.5, 0.75, 1],
            [0, 0.5, 0.5, 0.75, 1],
            [0, 0.5, 0.5, 0.75, 1],
            [0, 0.5, 0.5, 0.75, 1],
        ]
    ).T
    wdud_val = wdud(dist)
    # calculated using wolfram alpha:
    expected = 0.065 + 0.01625 + 0.02125
    assert_almost_equal(wdud_val, expected, decimal=4)


def test_wdud_dimension_error():
    """Test wdud method raises error when input has incorrect dimensions."""
    arr = np.zeros((2, 2, 2))
    assert_raises(ValueError, wdud, arr)


def test_wdud_normalization_error():
    """Test wdud method raises error when normalization fails."""
    assert_raises(ValueError, wdud, sample8)


def test_wdud_warning_normalization():
    """Test wdud method gives warning when normalization fails."""
    warning_message = (
        "Some of the features are constant which will cause the normalization to fail. "
        + "Now removing them."
    )
    with pytest.warns() as record:
        wdud(sample6)

    # check that the message matches
    assert record[0].message.args[0] == warning_message


def test_hypersphere_overlap_of_subset_with_only_corners_and_center():
    """Test the hypersphere overlap method with predefined matrix."""
    corner_pts = np.array([[0.0, 0.0], [0.0, 1.0], [1.0, 0.0], [1.0, 1.0]])
    # Many duplicate pts cause r_0 to be much smaller than 1.0,
    #   which is required due to normalization of the feature space
    centers_pts = np.array([[0.5, 0.5]] * (100 - 4))
    pts = np.vstack((corner_pts, centers_pts))

    # Overlap should be all coming from the centers
    expected_overlap = 100.0 * 96 * 95 * 0.5
    # The edge penalty should all be from the corner pts
    lam = 1.0 / 2.0  # Default lambda chosen from paper.
    expected_edge = lam * 4.0
    expected = expected_overlap + expected_edge
    true = hypersphere_overlap_of_subset(pts, pts)
    assert_almost_equal(true, expected)


def test_hypersphere_normalization_error():
    """Test the hypersphere overlap method raises error when normalization fails."""
    assert_raises(ValueError, hypersphere_overlap_of_subset, sample7, sample7)


def test_hypersphere_radius_warning():
    """Test the hypersphere overlap method gives warning when radius is too large."""
    corner_pts = np.array([[0.0, 0.0], [0.0, 1.0], [1.0, 0.0], [1.0, 1.0]])
    assert_warns(Warning, hypersphere_overlap_of_subset, corner_pts, corner_pts)


def test_gini_coefficient_of_non_diverse_set():
    """Test Gini coefficient of the least diverse set. Expected return is zero."""
    # Finger-prints where columns are all the same
    numb_molecules = 5
    numb_features = 10
    # Transpose so that the columns are all the same, note first made the rows all same
    single_fingerprint = list(np.random.choice([0, 1], size=(numb_features,)))
    finger_prints = np.array([single_fingerprint] * numb_molecules).T

    result = gini_coefficient(finger_prints)
    # Since they are all the same, then gini coefficient should be zero.
    assert_almost_equal(result, 0.0, decimal=8)


def test_gini_coefficient_non_binary_error():
    """Test Gini coefficient error when input is not binary."""
    assert_raises(ValueError, gini_coefficient, np.array([[7, 0], [2, 1]]))


def test_gini_coefficient_dimension_error():
    """Test Gini coefficient error when input has incorrect dimensions."""
    assert_raises(ValueError, gini_coefficient, np.array([1, 0, 0, 0]))


def test_gini_coefficient_of_most_diverse_set():
    """Test Gini coefficient of the most diverse set."""
    #  Finger-prints where one feature has more `wealth` than all others.
    #  Note: Transpose is done so one column has all ones.
    finger_prints = np.array(
        [
            [1, 1, 1, 1, 1, 1, 1],
        ]
        + [[0, 0, 0, 0, 0, 0, 0]] * 100000
    ).T
    result = gini_coefficient(finger_prints)
    # Since they are all the same, then gini coefficient should be zero.
    assert_almost_equal(result, 1.0, decimal=4)


def test_gini_coefficient_with_alternative_definition():
    """Test Gini coefficient with alternative definition."""
    # Finger-prints where they are all different
    numb_features = 4
    finger_prints = np.array([[1, 1, 1, 1], [0, 1, 1, 1], [0, 0, 1, 1], [0, 0, 0, 1]])
    result = gini_coefficient(finger_prints)

    # Alternative definition from wikipedia
    b = numb_features + 1
    desired = (
        numb_features + 1 - 2 * ((b - 1) + (b - 2) * 2 + (b - 3) * 3 + (b - 4) * 4) / (10)
    ) / numb_features
    assert_almost_equal(result, desired)


def test_nearest_average_tanimoto_bit():
    """Test the nearest_average_tanimoto function with binary input."""
    nat = nearest_average_tanimoto(sample2)
    shortest_tani = [0.3333333, 0.3333333, 0, 0]
    average = np.average(shortest_tani)
    assert_almost_equal(nat, average)


def test_nearest_average_tanimoto():
    """Test the nearest_average_tanimoto function with non-binary input."""
    nat = nearest_average_tanimoto(sample3)
    shortest_tani = [(11 / 19), (11 / 19)]
    average = np.average(shortest_tani)
    assert_equal(nat, average)


def test_nearest_average_tanimoto_3_x_3():
    """Testpyth the nearest_average_tanimoto function with a 3x3 matrix."""
    # all unequal distances b/w points
    x = np.array([[0, 1, 2], [3, 4, 5], [4, 5, 6]])
    nat_x = nearest_average_tanimoto(x)
    avg_x = 0.749718574108818
    assert_equal(nat_x, avg_x)
    # one point equidistant from the other two
    y = np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])
    nat_y = nearest_average_tanimoto(y)
    avg_y = 0.4813295920569825
    assert_equal(nat_y, avg_y)
    # all points equidistant
    z = np.array([[0, 1, 2], [1, 2, 0], [2, 0, 1]])
    nat_z = nearest_average_tanimoto(z)
    avg_z = 0.25
    assert_equal(nat_z, avg_z)


def test_nearest_average_tanimoto_nonsquare():
    """Test the nearest_average_tanimoto function with non-binary input"""
    x = np.array([[3.5, 4.0, 10.5, 0.5], [1.25, 4.0, 7.0, 0.1], [0.0, 0.0, 0.0, 0.0]])
    # nearest neighbor of sample 0, 1, and 2 are sample 1, 0, and 1, respectively.
    expected = np.average(
        [
            np.sum(x[0] * x[1]) / (np.sum(x[0] ** 2) + np.sum(x[1] ** 2) - np.sum(x[0] * x[1])),
            np.sum(x[1] * x[0]) / (np.sum(x[1] ** 2) + np.sum(x[0] ** 2) - np.sum(x[1] * x[0])),
            np.sum(x[2] * x[1]) / (np.sum(x[2] ** 2) + np.sum(x[1] ** 2) - np.sum(x[2] * x[1])),
        ]
    )
    assert_equal(nearest_average_tanimoto(x), expected)



# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Test similarity.py."""

import ast
import csv
import importlib

import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_equal, assert_raises

from selector.measures.similarity import (
    modified_tanimoto,
    pairwise_similarity_bit,
    scaled_similarity_matrix,
    tanimoto,
)
from selector.methods.similarity import NSimilarity, SimilarityIndex
from selector.methods.tests.common import get_data_file_path


def test_pairwise_similarity_bit_raises():
    # check raised error for input feature matrix that is not 2D
    assert_raises(ValueError, pairwise_similarity_bit, np.random.random(5), "tanimoto")
    assert_raises(ValueError, pairwise_similarity_bit, np.random.random((2, 3, 4)), "tanimoto")
    # check raised error for not-available method
    assert_raises(ValueError, pairwise_similarity_bit, np.random.random((5, 1)), "tan")
    assert_raises(ValueError, pairwise_similarity_bit, np.random.random((5, 1)), tanimoto)


def test_tanimoto_raises():
    # check raised error when a or b is not 1D
    assert_raises(ValueError, tanimoto, np.random.random((1, 5)), np.random.random(5))
    assert_raises(ValueError, tanimoto, np.random.random(3), np.random.random((1, 4)))
    assert_raises(ValueError, tanimoto, np.random.random(4), np.random.random((3, 4)))
    assert_raises(ValueError, tanimoto, np.random.random((3, 3)), np.random.random((2, 3)))
    # check raised error when a and b don't have the same length
    assert_raises(ValueError, tanimoto, np.random.random(3), np.random.random(5))
    assert_raises(ValueError, tanimoto, np.random.random(20), np.random.random(10))


def test_tanimoto():
    """Test the tanimoto function on one pair of points."""
    a = np.array([2, 0, 1])
    b = np.array([2, 0, 0])
    # expected = (2*2 + 0*0 + 1*0) / (2**2 + 1 + 2**2 - 2*2)
    assert_equal(tanimoto(a, b), 4 / (5 + 4 - 4))


def test_tanimoto_bitstring():
    """Test the tanimoto function on one pair of points."""
    a = np.array([0, 0, 0, 1, 0, 1, 1])
    b = np.array([1, 1, 0, 0, 0, 1, 1])
    assert_equal(tanimoto(a, b), 2 / 5)


def test_tanimoto_matrix():
    """Testing the tanimoto function with predefined feature matrix."""
    x = np.array([[1, 4], [3, 2]])
    s = pairwise_similarity_bit(x, "tanimoto")
    expected = np.array([[1, (11 / 19)], [(11 / 19), 1]])
    assert_equal(s, expected)


def test_modified_tanimoto():
    a = np.array([1, 1, 0, 0, 1])
    b = np.array([0, 0, 0, 0, 1])
    expected = (1.6 / 9) + (1.4 / 6)
    assert_equal(modified_tanimoto(a, b), expected)


def test_modified_tanimoto_all_ones():
    """Test the modified tanimoto function when input is all '1' bits"""
    a = np.array([1, 1, 1, 1, 1])
    assert_equal(modified_tanimoto(a, a), 1)


def test_modified_tanimoto_all_zeroes():
    """Test the modified tanimoto function when input is all '0' bits"""
    a = np.zeros(5)
    assert_equal(modified_tanimoto(a, a), 1)


def test_modified_tanimoto_dimension_error():
    """Test modified tanimoto raises error when input has incorrect dimension."""
    a = np.zeros([7, 5])
    b = np.zeros(5)
    assert_raises(ValueError, modified_tanimoto, a, b)
    assert_raises(ValueError, modified_tanimoto, b, a)
    assert_raises(ValueError, modified_tanimoto, np.ones(3), np.ones(5))


def test_modified_tanimoto_matrix():
    """Testing the modified tanimoto function with predefined feature matrix."""
    x = np.array([[1, 0, 1], [0, 1, 1]])
    s = pairwise_similarity_bit(x, "modified_tanimoto")
    expected = np.array([[1, (4 / 27)], [(4 / 27), 1]])
    assert_equal(s, expected)


def test_scaled_similarity_matrix():
    """Testing scaled similarity matrix function with a predefined similarity matrix."""
    X = np.array([[4, 2, 3], [0.5, 1, 1.5], [3, 3, 9]])
    s = scaled_similarity_matrix(X)
    expected = np.array([[1, 1, 0.5], [0.25, 1, 0.5], [0.5, 1, 1]])
    assert_equal(s, expected)


def test_scaled_similarity_matrix_dimension_error():
    """Test scaled similarity matrix raises error when input has incorrect dimension."""
    X = np.array([[1, 0.8, 0.65], [0.8, 1, 0.47]])
    assert_raises(ValueError, scaled_similarity_matrix, X)


def test_scaled_similarity_matrix_diagonal_zero_error():
    """Test scaled similarity matrix raises error when diagonal element is zero."""
    X = np.array([[1, 0.8, 0.65], [0.8, 0, 0.47], [0.65, 0.47, 1]])
    assert_raises(ValueError, scaled_similarity_matrix, X)


def test_scaled_similarity_matrix_element_negative_error():
    """Test scaled similarity matrix raises error when any element is negative."""
    X = np.array([[1, 0.8, 0.65], [-0.8, 1, 0.47], [0.65, 0.47, 1]])
    assert_raises(ValueError, scaled_similarity_matrix, X)


def test_SimilarityIndex_init_raises():
    """Test the SimilarityIndex class for raised errors (initialization)."""
    # check raised error wrong similarity index name
    with pytest.raises(ValueError):
        SimilarityIndex(method="esim", sim_index="ttt")
    # check raised error wrong c_threshold - invalid string value
    with pytest.raises(ValueError):
        SimilarityIndex(method="esim", c_threshold="ttt")
    # check raised error wrong c_threshold - invalid type (not int)
    with pytest.raises(ValueError):
        SimilarityIndex(method="esim", c_threshold=1.1)


def test_SimilarityIndex_calculate_counters_raises():
    sim_idx = SimilarityIndex(method="esim")

    # check raised error wrong data type
    with pytest.raises(TypeError):
        sim_idx._calculate_counters(X=[1, 2, 3])

    # check raised error - no n_objects with data of length 1
    with pytest.raises(ValueError):
        sim_idx._calculate_counters(X=np.array([1, 2, 3]))

    # check raised error - c_threshold bigger than n_objects
    sim_idx = SimilarityIndex(method="esim", c_threshold=3)
    with pytest.raises(ValueError):
        sim_idx._calculate_counters(X=np.array([[1, 2, 3], [4, 5, 6]]))

    # check raised error - invalid c_threshold string value
    sim_idx = SimilarityIndex(method="esim")
    sim_idx.c_threshold = "ttt"
    with pytest.raises(ValueError):
        sim_idx._calculate_counters(X=np.array([[1, 2, 3], [4, 5, 6]]))

    # check raised error - invalid weight factor string value
    sim_idx = SimilarityIndex(method="esim")
    sim_idx.w_factor = "ttt"
    with pytest.raises(ValueError):
        sim_idx._calculate_counters(X=np.array([[1, 2, 3], [4, 5, 6]]))


def test_SimilarityIndex_call_raises():
    """Test the SimilarityIndex class for raised errors (call)."""
    sim_idx = SimilarityIndex()
    # check raised error wrong data type
    with pytest.raises(TypeError):
        sim_idx(X=[1, 2, 3])
    # check raised error - no n_objects with data of length 1
    with pytest.raises(ValueError):
        sim_idx(X=np.array([1, 2, 3]))


def test_SimilarityIndex_calculate_medoid_raises():
    """Test the SimilarityIndex class for raised errors (calculate_medoid)."""
    sim_idx = NSimilarity()

    # check raised error wrong data type
    with pytest.raises(TypeError):
        sim_idx.calculate_medoid(X=[1, 2, 3])

    # check raised error - no medoid with one dimensional data
    with pytest.raises(ValueError):
        sim_idx.calculate_medoid(X=np.array([1, 2, 3]))

    # check raised error - no medoid with less than three samples
    with pytest.raises(ValueError):
        sim_idx.calculate_medoid(X=np.array([[1, 2, 3], [4, 5, 6]]))

    # check raised error - c_threshold bigger than n_objects
    sim_idx = NSimilarity(method="esim", c_threshold=4)
    with pytest.raises(ValueError):
        sim_idx.calculate_medoid(X=np.array([[1, 2, 3], [4, 5, 6], [6, 7, 8]]))

    # check raised error - c_total and data have different number of columns
    sim_idx = NSimilarity()
    with pytest.raises(ValueError):
        sim_idx.calculate_medoid(
            X=np.array([[1, 2, 3], [4, 5, 6], [6, 7, 8]]), c_total=np.array([1, 2, 3, 4])
        )


def test_SimilarityIndex_calculate_outlier_raises():
    """Test the SimilarityIndex class for raised errors (calculate_outlier)."""
    sim_idx = NSimilarity()

    # check raised error wrong data type
    with pytest.raises(TypeError):
        sim_idx.calculate_outlier(X=[1, 2, 3])

    # check raised error - no medoid with one dimensional data
    with pytest.raises(ValueError):
        sim_idx.calculate_outlier(X=np.array([1, 2, 3]))

    # check raised error - no medoid with less than three samples
    with pytest.raises(ValueError):
        sim_idx.calculate_outlier(X=np.array([[1, 2, 3], [4, 5, 6]]))

    # check raised error - c_threshold bigger than n_objects
    sim_idx = NSimilarity(method="esim", c_threshold=4)
    with pytest.raises(ValueError):
        sim_idx.calculate_outlier(X=np.array([[1, 2, 3], [4, 5, 6], [6, 7, 8]]))

    # check raised error - c_total and data have different number of columns
    sim_idx = NSimilarity()
    with pytest.raises(ValueError):
        sim_idx.calculate_outlier(
            X=np.array([[1, 2, 3], [4, 5, 6], [6, 7, 8]]), c_total=np.array([1, 2, 3, 4])
        )


def test_NSimilarity_init_raises():
    """Test the NSimilarity class for raised errors (initialization)."""
    # check raised error wrong similarity index name
    with pytest.raises(ValueError):
        NSimilarity(similarity_index="ttt")
    # check raised error wrong c_threshold - invalid string value
    with pytest.raises(ValueError):
        NSimilarity(method="esim", c_threshold="ttt")
    # check raised error wrong c_threshold - invalid type (not int)
    with pytest.raises(ValueError):
        NSimilarity(method="esim", c_threshold=1.1)


def test_NSimilarity_get_new_index_raises():
    """Test the NSimilarity class for raised errors (get_new_index)."""
    # check that data is not scaled between 0 and 1
    with pytest.raises(ValueError):
        NSimilarity()._get_new_index(
            X=np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
            num_selected=1,
            selected_condensed=np.array([1, 2, 3]),
            select_from=np.array([1, 2, 3]),
        )


def test_NSimilarity_select_from_cluster_raises():
    """Test the NSimilarity class for raised errors (select_from_cluster)."""
    data_array = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    cluster_ids = np.array([0, 1, 2])
    with pytest.raises(ValueError):
        # select more samples than the number of samples in the cluster
        NSimilarity().select_from_cluster(X=data_array, size=5)
    with pytest.raises(ValueError):
        # start from sample index that is not in the cluster
        NSimilarity().select_from_cluster(X=data_array, size=2, start=[4])
    with pytest.raises(ValueError):
        # start from sample index that is not in the cluster
        NSimilarity().select_from_cluster(X=data_array, size=2, start=[4])
    with pytest.raises(ValueError):
        # start from invalid string value of start
        NSimilarity().select_from_cluster(X=data_array, size=2, start="ttt")
    with pytest.raises(ValueError):
        # start from invalid type of start
        NSimilarity().select_from_cluster(X=data_array, size=2, start=[1.2])
    with pytest.raises(ValueError):
        # try to data not scaled between 0 and 1
        NSimilarity(preprocess_data=False).select_from_cluster(X=data_array, size=2)
    with pytest.raises(ValueError):
        # try to use starting index that is not in the cluster
        NSimilarity().select_from_cluster(X=data_array, size=2, start=[4], labels=cluster_ids)
    with pytest.raises(ValueError):
        # try to use invalid starting index
        NSimilarity().select_from_cluster(X=data_array, size=2, start=4.2, labels=cluster_ids)


# --------------------------------------------------------------------------------------------- #
# Tests for the function results of the SimilarityIndex and NSimilarity classes.
# --------------------------------------------------------------------------------------------- #
# The following part tests the results of the SimilarityIndex and NSimilarity classes methods
# for a set of binary data. The proper results for the tests are known in advance.

# Tests for binary data.
# --------------------------------------------------------------------------------------------- #

# --------------------------------------------------------------------------------------------- #
# Section of the test for the SimilarityIndex class using binary data and esim method.
# --------------------------------------------------------------------------------------------- #


def _get_binary_data():
    """Returns a list of binary strings.

    The proper results for the tests are known in advance.

    Returns
    -------
    list of lists
        A list of binary objects lists.
    """

    # Binary data to perform the tests, it contains 100 lists of 12 elements each
    # The proper results for the tests are known in advance.
    data = np.array(
        [
            [0, 0, 0, 1, 0, 1, 1, 1, 0, 1, 1, 1],
            [1, 1, 0, 1, 1, 1, 0, 0, 1, 0, 1, 1],
            [1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0],
            [0, 1, 1, 0, 1, 1, 1, 1, 0, 1, 0, 1],
            [0, 1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1],
            [1, 0, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0],
            [1, 1, 0, 1, 1, 0, 0, 1, 0, 0, 1, 1],
            [0, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0],
            [0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1],
            [0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 1],
            [1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0],
            [1, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 1],
            [1, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0],
            [1, 1, 0, 0, 1, 0, 1, 0, 0, 0, 1, 1],
            [1, 1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 0],
            [1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1],
            [1, 0, 1, 1, 0, 0, 1, 0, 1, 1, 1, 0],
            [1, 1, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0],
            [1, 1, 1, 1, 0, 1, 1, 1, 0, 0, 1, 1],
            [0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1],
            [1, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0],
            [0, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 1],
            [0, 0, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0],
            [1, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 1],
            [1, 0, 1, 0, 0, 1, 1, 0, 0, 0, 1, 1],
            [1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1],
            [1, 1, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
            [1, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0],
            [0, 0, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0],
            [1, 0, 1, 1, 0, 1, 0, 0, 0, 1, 1, 1],
            [0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0],
            [0, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 1],
            [0, 0, 1, 1, 1, 0, 1, 0, 1, 1, 0, 0],
            [0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1],
            [1, 1, 0, 1, 1, 0, 0, 0, 1, 0, 1, 0],
            [1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 0],
            [0, 1, 1, 1, 0, 1, 0, 1, 1, 1, 1, 0],
            [1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1],
            [1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0],
            [0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1],
            [1, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 0],
            [1, 0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0],
            [0, 1, 0, 0, 0, 1, 0, 1, 1, 1, 1, 0],
            [0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 0, 0],
            [1, 0, 1, 1, 0, 1, 0, 1, 0, 0, 1, 1],
            [1, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 1],
            [0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 1, 0],
            [0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0],
            [1, 0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 1],
            [0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0],
            [1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 0],
            [0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 0],
            [0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 1],
            [0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0],
            [0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0],
            [1, 0, 0, 0, 1, 0, 1, 1, 1, 1, 1, 1],
            [0, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 0],
            [0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0],
            [1, 1, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0],
            [1, 1, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0],
            [0, 0, 0, 1, 1, 0, 1, 1, 1, 1, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0],
            [0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 0],
            [0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 1, 0],
            [1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0],
            [1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 0, 0],
            [0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1],
            [0, 1, 1, 0, 0, 1, 0, 1, 1, 1, 0, 0],
            [1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 1, 1],
            [0, 1, 0, 1, 0, 1, 0, 0, 1, 1, 0, 1],
            [1, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0],
            [1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 0, 1],
            [1, 1, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0],
            [1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0],
            [1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1],
            [0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0],
            [1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0],
            [1, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0],
            [1, 0, 1, 1, 1, 1, 0, 1, 0, 0, 0, 1],
            [1, 1, 0, 1, 0, 0, 1, 0, 0, 1, 1, 0],
            [0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 1, 1],
            [0, 1, 1, 0, 0, 0, 1, 0, 1, 1, 0, 1],
            [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0],
            [0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0],
            [0, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1],
            [1, 1, 0, 1, 1, 1, 1, 1, 1, 0, 0, 1],
            [0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 1, 1],
            [0, 0, 1, 0, 1, 1, 1, 0, 1, 1, 0, 0],
            [0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 1],
            [1, 1, 0, 1, 0, 0, 1, 1, 0, 1, 0, 0],
            [1, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0],
            [1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 0, 1],
            [1, 1, 0, 0, 0, 1, 0, 1, 0, 0, 1, 1],
            [1, 1, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0],
            [0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1],
            [0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1],
            [0, 1, 1, 1, 0, 1, 0, 1, 1, 0, 1, 1],
        ]
    )

    return data


def _get_ref_similarity_esim_dict():
    """Returns a dictionary with the reference values for the similarity indexes.

    The proper results for the tests are known in advance and are stored in a dictionary.

    Returns
    -------
    dict
        A dictionary with the reference values for the similarity indexes. The dictionary has the
        following structure:
            {w_factor: {c_threshold: {n_ary: value}}} where:
                w_factor: the weight factor for the similarity index
                c_threshold: the threshold value for the similarity index
    """

    # Reference values for the similarity index.
    ref_similarity_binary = {
        "fraction": {
            "None": {
                "AC": 0.1597619835634367,
                "BUB": 0.06891469248218901,
                "CT1": 0.21594387881973748,
                "CT2": 0.12273565515991067,
                "CT3": 0.09010381444416192,
                "CT4": 0.1435977860207451,
                "Fai": 0.041666666666666664,
                "Gle": 0.07428571428571429,
                "Ja": 0.078,
                "Ja0": 0.06529411764705882,
                "JT": 0.065,
                "RT": 0.05692307692307692,
                "RR": 0.021666666666666667,
                "SM": 0.06166666666666667,
                "SS1": 0.052000000000000005,
                "SS2": 0.06434782608695652,
            },
            "dissimilar": {
                "AC": 0.0,
                "BUB": 0.0,
                "CT1": 0.0,
                "CT2": 0.0,
                "CT3": 0.0,
                "CT4": 0.0,
                "Fai": 0.0,
                "Gle": 0.0,
                "Ja": 0.0,
                "Ja0": 0.0,
                "JT": 0.0,
                "RT": 0.0,
                "RR": 0.0,
                "SM": 0.0,
                "SS1": 0.0,
                "SS2": 0.0,
            },
            5: {
                "AC": 0.13861833940391668,
                "BUB": 0.043638272469757794,
                "CT1": 0.173370214880788,
                "CT2": 0.02399965239873406,
                "CT3": 0.07108193238341105,
                "CT4": 0.0791812460476248,
                "Fai": 0.03166666666666667,
                "Gle": 0.04,
                "Ja": 0.05454545454545456,
                "Ja0": 0.084,
                "JT": 0.022222222222222223,
                "RT": 0.028000000000000004,
                "RR": 0.016666666666666666,
                "SM": 0.04666666666666667,
                "SS1": 0.011764705882352941,
                "SS2": 0.07,
            },
        },
        "power_3": {
            "None": {
                "AC": 1.5127728116149686e-20,
                "BUB": 7.90100730602568e-40,
                "CT1": 0.0,
                "CT2": 0.0,
                "CT3": 0.0,
                "CT4": 0.0,
                "Fai": 5.6422575986042295e-40,
                "Gle": 1.9329988216613272e-39,
                "Ja": 2.029648762744394e-39,
                "Ja0": 5.978755024266622e-40,
                "JT": 1.6913739689536614e-39,
                "RT": 5.212247969873465e-40,
                "RR": 5.6379132298455384e-40,
                "SM": 5.646601967362921e-40,
                "SS1": 1.3530991751629292e-39,
                "SS2": 5.892106400726526e-40,
            },
            "dissimilar": {
                "AC": 0.0,
                "BUB": 0.0,
                "CT1": 0.0,
                "CT2": 0.0,
                "CT3": 0.0,
                "CT4": 0.0,
                "Fai": 0.0,
                "Gle": 0.0,
                "Ja": 0.0,
                "Ja0": 0.0,
                "JT": 0.0,
                "RT": 0.0,
                "RR": 0.0,
                "SM": 0.0,
                "SS1": 0.0,
                "SS2": 0.0,
            },
            5: {
                "AC": 1.5127727667796592e-20,
                "BUB": 6.551485125947553e-40,
                "CT1": 0.0,
                "CT2": 0.0,
                "CT3": 0.0,
                "CT4": 0.0,
                "Fai": 5.642257358488985e-40,
                "Gle": 1.3530991402370754e-39,
                "Ja": 1.8451351912323755e-39,
                "Ja0": 1.0163882938782278e-39,
                "JT": 7.517217445761529e-40,
                "RT": 3.387960979594093e-40,
                "RR": 5.637913084321148e-40,
                "SM": 5.646601632656821e-40,
                "SS1": 3.979703353638457e-40,
                "SS2": 8.469902448985232e-40,
            },
        },
    }

    return ref_similarity_binary


def _get_ref_isim_dict():
    """Returns a dictionary with the reference values necessary for testing the isim method.

    The proper results for the tests are known in advance and are stored in a dictionary.

    Returns
    -------
    dict
        A dictionary with the reference values necessary for all the isim method tests. The
        dictionary has the following structure:
            {n_ary: { outlier: value, medoid: value, kval2si: {k: value}}} where:
                n_ary: the similarity index name used
                outlier: the index of the outlier sample
                medoid: the index of the medoid sample
                kval2si: a dictionary with the k value as key and the similarity index as value
    """
    isim_dict = {
        "AC": {
            "outlier": 49,
            "medoid": 96,
            "kval2si": {
                1: 0.49917475122243926,
                2: 0.5538883964836272,
                5: 0.5866150284585204,
                10: 0.5974286117693461,
            },
        },
        "BUB": {
            "outlier": 85,
            "medoid": 10,
            "kval2si": {
                1: 0.49397989672931264,
                2: 0.5819613020487422,
                5: 0.6335183244648649,
                10: 0.6502407823406905,
            },
        },
        "CT1": {
            "outlier": 49,
            "medoid": 96,
            "kval2si": {
                1: 0.9367065121244419,
                2: 0.9261583241959612,
                5: 0.9107923779571723,
                10: 0.902405671341302,
            },
        },
        "CT2": {
            "outlier": 49,
            "medoid": 96,
            "kval2si": {
                1: 0.06282178200501828,
                2: 0.12056333324739296,
                5: 0.19665579893673796,
                10: 0.23722036836219337,
            },
        },
        "CT3": {
            "outlier": 85,
            "medoid": 10,
            "kval2si": {
                1: 0.870243523865252,
                2: 0.82843367496858,
                5: 0.77434619420322,
                10: 0.7461466827474635,
            },
        },
        "CT4": {
            "outlier": 85,
            "medoid": 10,
            "kval2si": {
                1: 0.8945856132657759,
                2: 0.8706848775823933,
                5: 0.8375702929632953,
                10: 0.8197597232141359,
            },
        },
        "Fai": {
            "outlier": 85,
            "medoid": 10,
            "kval2si": {
                1: 0.36944444444444446,
                2: 0.4354778372150966,
                5: 0.47461264224551747,
                10: 0.4873954166765502,
            },
        },
        "Gle": {
            "outlier": 85,
            "medoid": 10,
            "kval2si": {
                1: 0.48934163365402755,
                2: 0.579696056605444,
                5: 0.6326571892686095,
                10: 0.6498200476852635,
            },
        },
        "Ja": {
            "outlier": 85,
            "medoid": 10,
            "kval2si": {
                1: 0.5897241588360109,
                2: 0.6741446014011255,
                5: 0.7209338061721255,
                10: 0.735695295519683,
            },
        },
        "Ja0": {
            "outlier": 49,
            "medoid": 96,
            "kval2si": {
                1: 0.7490265158538847,
                2: 0.8082737370566611,
                5: 0.8388469288245588,
                10: 0.8482013298830383,
            },
        },
        "JT": {
            "outlier": 85,
            "medoid": 10,
            "kval2si": {
                1: 0.3239260739260739,
                2: 0.4081492974102138,
                5: 0.4626909830536219,
                10: 0.48128402926677866,
            },
        },
        "RT": {
            "outlier": 49,
            "medoid": 96,
            "kval2si": {
                1: 0.3321820648822006,
                2: 0.41267273273100585,
                5: 0.4645381672067039,
                10: 0.48220837765022756,
            },
        },
        "RR": {
            "outlier": 85,
            "medoid": 10,
            "kval2si": {
                1: 0.2401851851851852,
                2: 0.2867117470582547,
                5: 0.3148435010613659,
                10: 0.3241287740619748,
            },
        },
        "SM": {
            "outlier": 49,
            "medoid": 96,
            "kval2si": {
                1: 0.4987037037037037,
                2: 0.5842439273719385,
                5: 0.6343817834296691,
                10: 0.6506620592911254,
            },
        },
        "SS1": {
            "outlier": 85,
            "medoid": 10,
            "kval2si": {
                1: 0.19326478915213827,
                2: 0.25639923187909175,
                5: 0.3009746107992553,
                10: 0.3169019346220607,
            },
        },
        "SS2": {
            "outlier": 49,
            "medoid": 96,
            "kval2si": {
                1: 0.6655134066477203,
                2: 0.7375681450029299,
                5: 0.7762957099270287,
                10: 0.7883649540846069,
            },
        },
    }
    return isim_dict


def _get_absolute_decimal_places_for_comparison(num1, num2, rtol=1e-4):
    """Calculate the absolute number of decimal places needed for comparison of two numbers.

    Calculate the absolute number of decimal places needed for assert_almost_equal of two numbers
    with a relative tolerance of rtol.

    Parameters
    ----------
    num1 : float
        First number.
    num2 : float
        Second number.
    rtol : float
        Relative tolerance.
    """
    max_num = max(abs(num1), abs(num2))

    # If both numbers are zero, return 5 decimal places just to avoid division by zero. Both numbers
    # are zero (and therefore equal) withing the machine precision.
    if max_num == 0:
        return 5

    # Calculate the dynamic tolerance based on the magnitudes of the numbers
    tol = rtol * max(abs(num1), abs(num2))

    return abs(int(np.log10(tol)))


# Parameter values for the tests
c_treshold_values = [None, "dissimilar", 5]
w_factor_values = ["fraction", "power_3"]
n_ary_values = [
    "AC",
    "BUB",
    "CT1",
    "CT2",
    "CT3",
    "CT4",
    "Fai",
    "Gle",
    "Ja",
    "Ja0",
    "JT",
    "RT",
    "RR",
    "SM",
    "SS1",
    "SS2",
]


@pytest.mark.parametrize("c_threshold", c_treshold_values)
@pytest.mark.parametrize("w_factor", w_factor_values)
@pytest.mark.parametrize("n_ary", n_ary_values)
def test_SimilarityIndex_esim_call(c_threshold, w_factor, n_ary):
    """Test the similarity index for binary data.

    Test the similarity index for binary data using the reference values and several combinations
    of parameters.

    Parameters
    ----------
    c_threshold : float
        The threshold value for the similarity index.
    w_factor : float
        The weight factor for the similarity index.
    n_ary : str
        The similarity index to use.
    """

    # get the binary data
    data, ref_similarity_binary = _get_binary_data(), _get_ref_similarity_esim_dict()

    # create instance of the class SimilarityIndex to test the similarity indexes for binary data
    sim_idx = SimilarityIndex(
        method="esim", sim_index=n_ary, c_threshold=c_threshold, w_factor=w_factor
    )

    # calculate the similarity index for the binary data
    sim_idx_value = sim_idx(data)

    # get the reference value for the similarity index
    if c_threshold is None:
        c_threshold = "None"
    ref_value = ref_similarity_binary[w_factor][c_threshold][n_ary]

    # calculate the absolute tolerance based on the relative tolerance and the numbers magnitude
    tol = _get_absolute_decimal_places_for_comparison(sim_idx_value, ref_value)

    # check that the calculated value is equal to the reference value
    assert_almost_equal(sim_idx_value, ref_value, decimal=tol)


# --------------------------------------------------------------------------------------------- #
# Section of the tests for selection of indexes functions using binary data and esim method.
# --------------------------------------------------------------------------------------------- #

# Many of the combinations of parameters for the tests are too stringent and the similarity of the
# samples is zero. These cases are not considered because will produce inconsistent results for the
# selection of the new indexes, and the calculation of the medoid and the outlier.

# Creating the possible combinations of the parameters for the tests. The combinations are stored
# in a list of tuples. Each tuple has the following structure: (w_factor, c_threshold, n_ary)
# where:
#   w_factor: the weight factor for the similarity index
#   c_threshold: the threshold value for the similarity index
#   n_ary: the similarity index to use


# The cases where the similarity of the samples is zero are not considered because these will be
# inconsistent with the reference values for the selection of the new indexes, and the calculation
# of the medoid and the outlier.
def _get_esim_test_parameters():
    """Returns the parameters for the tests.

    Returns the parameters for the tests. The parameters are stored in a list of tuples. Each tuple
    has the following structure: (w_factor, c_threshold, n_ary) where:
        w_factor: the weight factor for the similarity index
        c_threshold: the threshold value for the similarity index
        n_ary: the similarity index to use

    Returns
    -------
    list
        A list of tuples with the parameters for the tests. Each tuple has the following structure:
        (w_factor, c_threshold, n_ary) where:
            w_factor: the weight factor for the similarity index
            c_threshold: the threshold value for the similarity index
            n_ary: the similarity index to use
    """
    # The cases where the similarity of the samples is zero are not considered because these will be
    # inconsistent with the reference values for the selection of the new indexes, and the
    # calculation of the medoid and the outlier.
    test_parameters = []
    for w in w_factor_values:
        for c in c_treshold_values:
            # small hack to avoid using the None value as a key in the dictionary
            c_key = c
            if c is None:
                c_key = "None"
            for n in n_ary_values:
                # ignore the cases where the similarity of the samples less than two percent
                if not _get_ref_similarity_esim_dict()[w][c_key][n] < 0.02:
                    test_parameters.append((c, w, n))

    return test_parameters


parameters = _get_esim_test_parameters()


def _get_ref_medoid_esim_dict():
    """Returns dictionary with reference medoid index for binary data using esim method.

    The proper results for the tests are known in advance and are stored in a dictionary.

    The dictionary has the following structure:
        {w_factor: {c_threshold: {n_ary: value}}} where:
            w_factor: the weight factor for the similarity index
            c_threshold: the threshold value for the similarity index

    The medoid values are stored for all possible combinations of the following parameter values:
        - w_factor: None, dissimilar, 5
        - c_threshold: fraction, power_3
        - n_ary: AC, BUB, CT1, CT2, CT3, CT4, Fai, Gle, Ja, Ja0, JT, RT, RR, SM, SS1, SS2


    Returns
    -------
    dict
        A dictionary with the reference values for the medoid. The dictionary has the
        following structure:
            {w_factor: {c_threshold: {n_ary: value}}} where:
                w_factor: the weight factor for the similarity index
                c_threshold: the threshold value for the similarity index
    """
    medoid_ref_dict = {
        "None": {
            "fraction": {
                "AC": 96,
                "BUB": 69,
                "CT1": 96,
                "CT2": 80,
                "CT3": 1,
                "CT4": 23,
                "Fai": 96,
                "Gle": 69,
                "Ja": 69,
                "Ja0": 50,
                "JT": 69,
                "RT": 96,
                "RR": 1,
                "SM": 96,
                "SS1": 23,
                "SS2": 50,
            },
            "power_3": {
                "AC": 96,
                "BUB": 48,
                "CT1": 0,
                "CT2": 0,
                "CT3": 0,
                "CT4": 0,
                "Fai": 96,
                "Gle": 69,
                "Ja": 69,
                "Ja0": 50,
                "JT": 69,
                "RT": 96,
                "RR": 1,
                "SM": 96,
                "SS1": 69,
                "SS2": 50,
            },
        },
        "dissimilar": {
            "fraction": {
                "AC": 0,
                "BUB": 0,
                "CT1": 0,
                "CT2": 0,
                "CT3": 0,
                "CT4": 0,
                "Fai": 0,
                "Gle": 0,
                "Ja": 0,
                "Ja0": 0,
                "JT": 0,
                "RT": 0,
                "RR": 0,
                "SM": 0,
                "SS1": 0,
                "SS2": 0,
            },
            "power_3": {
                "AC": 0,
                "BUB": 0,
                "CT1": 0,
                "CT2": 0,
                "CT3": 0,
                "CT4": 0,
                "Fai": 0,
                "Gle": 0,
                "Ja": 0,
                "Ja0": 0,
                "JT": 0,
                "RT": 0,
                "RR": 0,
                "SM": 0,
                "SS1": 0,
                "SS2": 0,
            },
        },
        5: {
            "fraction": {
                "AC": 39,
                "BUB": 39,
                "CT1": 39,
                "CT2": 96,
                "CT3": 0,
                "CT4": 0,
                "Fai": 39,
                "Gle": 0,
                "Ja": 0,
                "Ja0": 39,
                "JT": 0,
                "RT": 39,
                "RR": 0,
                "SM": 39,
                "SS1": 0,
                "SS2": 39,
            },
            "power_3": {
                "AC": 39,
                "BUB": 39,
                "CT1": 0,
                "CT2": 0,
                "CT3": 0,
                "CT4": 0,
                "Fai": 39,
                "Gle": 0,
                "Ja": 0,
                "Ja0": 39,
                "JT": 0,
                "RT": 39,
                "RR": 0,
                "SM": 39,
                "SS1": 0,
                "SS2": 39,
            },
        },
    }

    return medoid_ref_dict


@pytest.mark.parametrize("c_threshold, w_factor, n_ary", parameters)
def test_calculate_medoid_esim(c_threshold, w_factor, n_ary):
    """Test the function to calculate the medoid for binary data using esim method.

    Test the function to calculate the medoid for binary data using the reference values and several
    combinations of parameters. The reference values are obtained using the function
    _get_ref_medoid_dict().

    Parameters
    ----------
    c_threshold : float
        The threshold value for the similarity index.
    w_factor : float
        The weight factor for the similarity index.
    n_ary : str
        The similarity index to use.
    """

    # get the reference binary data
    data = _get_binary_data()
    # get the reference value for the medoid
    ref_medoid_dict = _get_ref_medoid_esim_dict()

    # small hack to avoid using the None value as a key in the dictionary
    c_threshold_key = c_threshold
    if c_threshold is None:
        c_threshold_key = "None"

    ref_medoid = ref_medoid_dict[c_threshold_key][w_factor][n_ary]

    # calculate the medoid for the binary data
    sim_idx = NSimilarity(
        method="esim", similarity_index=n_ary, c_threshold=c_threshold, w_factor=w_factor
    )
    medoid = sim_idx.calculate_medoid(X=data)

    # check that the calculated medoid is equal to the reference medoid
    assert_equal(medoid, ref_medoid)


# results from the reference code
def _get_ref_outlier_esim_dict():
    """Returns dictionary with reference outlier index for binary data using esim method.

    The proper results for the tests are known in advance and are stored in a dictionary.

        The dictionary has the following structure:
        {w_factor: {c_threshold: {n_ary: value}}} where:
            w_factor: the weight factor for the similarity index
            c_threshold: the threshold value for the similarity index

    The outlier values are stored for all possible combinations of the following parameter values:
        - w_factor: None, dissimilar, 5
        - c_threshold: fraction, power_3
        - n_ary: AC, BUB, CT1, CT2, CT3, CT4, Fai, Gle, Ja, Ja0, JT, RT, RR, SM, SS1, SS2


    Returns
    -------
    dict
        A dictionary with the reference values for the outlier. The dictionary has the
        following structure:
            {w_factor: {c_threshold: {n_ary: value}}} where:
                w_factor: the weight factor for the similarity index
                c_threshold: the threshold value for the similarity index
    """
    oulier_ref_dict = {
        "None": {
            "fraction": {
                "AC": 4,
                "BUB": 34,
                "CT1": 4,
                "CT2": 4,
                "CT3": 8,
                "CT4": 34,
                "Fai": 34,
                "Gle": 34,
                "Ja": 34,
                "Ja0": 49,
                "JT": 34,
                "RT": 4,
                "RR": 8,
                "SM": 4,
                "SS1": 34,
                "SS2": 49,
            },
            "power_3": {
                "AC": 49,
                "BUB": 57,
                "CT1": 0,
                "CT2": 0,
                "CT3": 0,
                "CT4": 0,
                "Fai": 49,
                "Gle": 34,
                "Ja": 16,
                "Ja0": 80,
                "JT": 34,
                "RT": 4,
                "RR": 8,
                "SM": 49,
                "SS1": 34,
                "SS2": 80,
            },
        },
        "dissimilar": {
            "fraction": {
                "AC": 0,
                "BUB": 0,
                "CT1": 0,
                "CT2": 0,
                "CT3": 0,
                "CT4": 0,
                "Fai": 0,
                "Gle": 0,
                "Ja": 0,
                "Ja0": 0,
                "JT": 0,
                "RT": 0,
                "RR": 0,
                "SM": 0,
                "SS1": 0,
                "SS2": 0,
            },
            "power_3": {
                "AC": 0,
                "BUB": 0,
                "CT1": 0,
                "CT2": 0,
                "CT3": 0,
                "CT4": 0,
                "Fai": 0,
                "Gle": 0,
                "Ja": 0,
                "Ja0": 0,
                "JT": 0,
                "RT": 0,
                "RR": 0,
                "SM": 0,
                "SS1": 0,
                "SS2": 0,
            },
        },
        5: {
            "fraction": {
                "AC": 49,
                "BUB": 49,
                "CT1": 49,
                "CT2": 57,
                "CT3": 2,
                "CT4": 2,
                "Fai": 49,
                "Gle": 2,
                "Ja": 2,
                "Ja0": 49,
                "JT": 2,
                "RT": 49,
                "RR": 2,
                "SM": 49,
                "SS1": 2,
                "SS2": 49,
            },
            "power_3": {
                "AC": 49,
                "BUB": 49,
                "CT1": 0,
                "CT2": 0,
                "CT3": 0,
                "CT4": 0,
                "Fai": 49,
                "Gle": 2,
                "Ja": 2,
                "Ja0": 49,
                "JT": 2,
                "RT": 49,
                "RR": 2,
                "SM": 49,
                "SS1": 2,
                "SS2": 49,
            },
        },
    }

    return oulier_ref_dict


@pytest.mark.parametrize("c_threshold, w_factor, n_ary", parameters)
def test_calculate_outlier_esim(c_threshold, w_factor, n_ary):
    """Test the function to calculate the outlier for binary data using the esim method.

    Test the function to calculate the outlier for binary data using the reference values and
    several combinations of parameters. The reference values are obtained using the function
    _get_ref_outlier_dict().

    Parameters
    ----------
    c_threshold : float
        The threshold value for the similarity index.
    w_factor : float
        The weight factor for the similarity index.
    n_ary : str
        The similarity index to use.
    """

    # get the reference binary data
    data = _get_binary_data()
    # get the reference value for the outlier
    ref_outlier_dict = _get_ref_outlier_esim_dict()

    # small hack to avoid using the None value as a key in the dictionary
    c_threshold_key = c_threshold
    if c_threshold is None:
        c_threshold_key = "None"

    ref_outlier = ref_outlier_dict[c_threshold_key][w_factor][n_ary]

    # calculate the outlier for the binary data
    sim_idx = NSimilarity(
        method="esim", similarity_index=n_ary, c_threshold=c_threshold, w_factor=w_factor
    )
    outlier = sim_idx.calculate_outlier(X=data)

    # check that the calculated outlier is equal to the reference outlier
    assert_equal(outlier, ref_outlier)


# --------------------------------------------------------------------------------------------- #
# Section of the tests for selection of indexes functions using binary data and isim method.
# --------------------------------------------------------------------------------------------- #
# test medoid selection
si_idcs = [
    "AC",
    "BUB",
    "CT1",
    "CT2",
    "CT3",
    "CT4",
    "Fai",
    "Gle",
    "Ja",
    "Ja0",
    "JT",
    "RT",
    "RR",
    "SM",
    "SS1",
    "SS2",
]


@pytest.mark.parametrize("sim_idx", si_idcs)
def test_calculate_medoid_isim(sim_idx):
    data = _get_binary_data()
    isim_dict = _get_ref_isim_dict()
    ref_medoid = isim_dict[sim_idx]["medoid"]
    sim_idx = NSimilarity(method="isim", similarity_index=sim_idx)
    medoid = sim_idx.calculate_medoid(X=data)
    assert_equal(medoid, ref_medoid)


# test outlier selection
@pytest.mark.parametrize("sim_idx", si_idcs)
def test_calculate_outlier_isim(sim_idx):
    data = _get_binary_data()
    isim_dict = _get_ref_isim_dict()
    ref_medoid = isim_dict[sim_idx]["outlier"]
    sim_idx = NSimilarity(method="isim", similarity_index=sim_idx)
    medoid = sim_idx.calculate_outlier(X=data)
    assert_equal(medoid, ref_medoid)


# --------------------------------------------------------------------------------------------- #
# Section of the test for the NSimilarity class using binary data and esim method.
# --------------------------------------------------------------------------------------------- #
def _get_ref_new_index_esim():
    """
    Returns reference data for testing the function to calculate the new index with the esim method.

    Returns the tuple (selected_points, new_indexes_dict) where selected_points is a list of
    selected samples given to the function (get_new_index) and new_indexes_dict is  a dictionary
    with the reference values for the new index to be selected based on the previously selected
    indices and the other function parameters.

    Returns
    -------
    tuple (list, dict)
        A tuple with the reference data for testing the function to calculate the new index. The
        tuple has the following structure:
            (selected_points, new_indexes_dict) where:
                selected_points: a list of selected samples given to the function (get_new_index)
                new_indexes_dict: a dictionary with the reference values for the new index to be
                    selected based on the previously selected indices and the other function
                    parameters. The dictionary has the following structure:
                        {w_factor: {c_threshold: {n_ary: value}}} where:
                            w_factor: the weight factor for the similarity index
                            c_threshold: the threshold value for the similarity index
                            n_ary: the similarity index to use

    """

    # The selected points to be given to the function (get_new_index)
    selected_samples = [0, 9, 19, 29, 39, 49, 59, 69, 79, 89, 99]
    # The reference values for the new index to be selected based on the previously selected
    # indices and the other function parameters.
    new_index_dict = {
        "None": {
            "fraction": {
                "AC": 62,
                "BUB": 62,
                "CT1": 62,
                "CT2": 62,
                "CT3": 2,
                "CT4": 44,
                "Fai": 62,
                "Gle": 44,
                "Ja": 44,
                "Ja0": 62,
                "JT": 44,
                "RT": 62,
                "RR": 2,
                "SM": 62,
                "SS1": 44,
                "SS2": 62,
            },
            "power_3": {
                "AC": 67,
                "BUB": 11,
                "CT1": 67,
                "CT2": 20,
                "CT3": 2,
                "CT4": 44,
                "Fai": 11,
                "Gle": 44,
                "Ja": 44,
                "Ja0": 11,
                "JT": 44,
                "RT": 67,
                "RR": 2,
                "SM": 67,
                "SS1": 44,
                "SS2": 11,
            },
        },
        5: {
            "fraction": {
                "AC": 2,
                "BUB": 2,
                "CT1": 2,
                "CT2": 2,
                "CT3": 2,
                "CT4": 2,
                "Fai": 2,
                "Gle": 2,
                "Ja": 2,
                "Ja0": 2,
                "JT": 2,
                "RT": 2,
                "RR": 2,
                "SM": 2,
                "SS1": 2,
                "SS2": 2,
            },
            "power_3": {
                "AC": 2,
                "BUB": 2,
                "CT1": 2,
                "CT2": 2,
                "CT3": 2,
                "CT4": 2,
                "Fai": 2,
                "Gle": 2,
                "Ja": 2,
                "Ja0": 2,
                "JT": 2,
                "RT": 2,
                "RR": 2,
                "SM": 2,
                "SS1": 2,
                "SS2": 2,
            },
        },
    }

    return selected_samples, new_index_dict


@pytest.mark.parametrize("c_threshold, w_factor, n_ary", parameters)
def test_get_new_index_esim(c_threshold, w_factor, n_ary):
    """Test the function get a new sample from the binary data.

    Test the function to get a new sample from the binary data using the reference values and
    several combinations of parameters. The reference values are obtained using the function
    _get_ref_new_index().

    Parameters
    ----------
    c_threshold : float
        The threshold value for the similarity index.
    w_factor : float
        The weight factor for the similarity index.
    n_ary : str
        The similarity index to use.
    """

    # get the reference binary data
    data = _get_binary_data()
    # get the reference value for the outlier
    selected_samples, ref_new_index_dict = _get_ref_new_index_esim()

    # columnwise sum of the selected samples
    selected_condensed_data = np.sum(np.take(data, selected_samples, axis=0), axis=0)
    # indices of the samples to select from
    select_from_n = [i for i in range(len(data)) if i not in selected_samples]
    # number of samples that are already selected
    n = len(selected_samples)

    # small hack to avoid using the None value as a key in the dictionary
    c_threshold_key = c_threshold
    if c_threshold is None:
        c_threshold_key = "None"

    ref_new_index = ref_new_index_dict[c_threshold_key][w_factor][n_ary]

    # calculate the outlier for the binary data
    nsi = NSimilarity(
        method="esim", similarity_index=n_ary, w_factor=w_factor, c_threshold=c_threshold
    )

    new_index = nsi._get_new_index(
        X=data,
        selected_condensed=selected_condensed_data,
        select_from=select_from_n,
        num_selected=n,
    )

    assert_equal(new_index, ref_new_index)


# --------------------------------------------------------------------------------------------- #
# Selecting a subset of the parameters generated for the tests
# --------------------------------------------------------------------------------------------- #

# remove cases where c_threshold is 5. This is not a valid case for the selection of the next point
# if the number of selected samples is less than 5
parameters = [x for x in parameters if not x[0] == 5]

# The start parameter can be a list of elements that need to be selected first.
# The case of the three elements included in a selection  1, 2, 3 is tested.
start_values = ["medoid", "outlier", [1, 2, 3]]
# sample size values to test
sample_size_values = [10, 20, 30]


# --------------------------------------------------------------------------------------------- #
# Get reference data for testing the selection of the diverse subset
# --------------------------------------------------------------------------------------------- #
# get reference selected data for esim method
def _get_selections_esim_ref_dict():
    """Returns a dictionary with the reference values for the selection of samples.

    The proper results for the tests are known in advance and are stored in a csv file.
    The file is read and the values are stored in a dictionary. The dictionary has the following
    structure:
        {w_factor: {c_threshold: {n_ary: {sample_size: {start: value}}}}} where:
            w_factor: the weight factor for the similarity index
            c_threshold: the threshold value for the similarity index
            n_ary: the similarity index to use
            sample_size: the number of samples to select
            start: the method to use to select the first(s) sample(s)
    """

    file_path = get_data_file_path("ref_esim_selection_data.csv")
    with open(file_path, encoding="utf-8") as file:
        reader = csv.reader(file, delimiter=";")
        next(reader)  # skip header
        # initialize the dictionary
        data_dict = {}

        for row in reader:
            # The first column is the c_threshold
            data_dict[row[0]] = data_dict.get(row[0], {})
            # The second column is the w_factor
            data_dict[row[0]][row[1]] = data_dict[row[0]].get(row[1], {})
            # The third column is the sample_size
            data_dict[row[0]][row[1]][row[2]] = data_dict[row[0]][row[1]].get(row[2], {})
            # The fourth column is the start_idx
            data_dict[row[0]][row[1]][row[2]][row[3]] = data_dict[row[0]][row[1]][row[2]].get(
                row[3], {}
            )
            # The fifth column stores the n_ary and the sixth column stores the reference value
            data_dict[row[0]][row[1]][row[2]][row[3]][row[4]] = ast.literal_eval(row[5])
    return data_dict


# test selected data for esim method (inv_order=1) and binary data against reference data
@pytest.mark.parametrize("c_threshold, w_factor, n_ary", parameters)
@pytest.mark.parametrize("sample_size", sample_size_values)
@pytest.mark.parametrize("start", start_values)
def test_NSimilarity_esim_select(c_threshold, w_factor, sample_size, n_ary, start):
    """
    Test the diversity selection methods based on similarity indexes for binary data.

    Parameters
    ----------
    c_threshold : float
        The threshold value for the similarity index.
    w_factor : float
        The weight factor for the similarity index.
    sample_size : int
        The number of molecules to select.
    n_ary : str
        The similarity index to use.
    start : str
        The method to use to select the first(s) sample(s).

    """

    # get the binary data
    data = _get_binary_data()
    # get the reference selected data
    reference_selected_data = _get_selections_esim_ref_dict()

    # create instance of the class SimilarityIndex to test the similarity indexes for binary data
    collector = NSimilarity(
        method="esim",
        similarity_index=n_ary,
        w_factor=w_factor,
        c_threshold=c_threshold,
        preprocess_data=False,
    )
    # select the diverse subset using the similarity index
    selected_data = collector.select_from_cluster(data, size=sample_size, start=start)

    # get the reference value for the similarity index
    # transform invalid keys to strings
    if c_threshold is None:
        c_threshold = "None"
    # transform sample_size to string (as it is used as a key in the dictionary)
    sample_size = str(sample_size)
    # lists of initial indexes are not valid keys in the dictionary (non hashable)
    if isinstance(start, list):
        start = str(start).replace(" ", "")
    # get the reference list as a string
    ref_list = reference_selected_data[c_threshold][w_factor][sample_size][start][n_ary]

    # check if the selected data has the same size as the reference data
    assert_equal(len(selected_data), len(ref_list))

    # check if the selected data is equal to the reference data
    assert all(x in ref_list for x in selected_data)



# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Common functions for test module."""

import numpy as np

try:
    from importlib_resources import path
except ImportError:
    from importlib.resources import path


def bit_cosine(a, b):
    """Compute dice coefficient.

    Parameters
    ----------
    a : array_like
        molecule A's features in bit string.
    b : array_like
        molecules B's features in bit string.

    Returns
    -------
    coeff : int
        dice coefficient for molecule A and B.
    """
    a_feat = np.count_nonzero(a)
    b_feat = np.count_nonzero(b)
    c = 0
    for idx, _ in enumerate(a):
        if a[idx] == b[idx] and a[idx] != 0:
            c += 1
    b_c = c / ((a_feat * b_feat) ** 0.5)
    return b_c


def bit_dice(a, b):
    """Compute dice coefficient.

    Parameters
    ----------
    a : array_like
        molecule A's features.
    b : array_like
        molecules B's features.

    Returns
    -------
    coeff : int
        dice coefficient for molecule A and B.
    """
    a_feat = np.count_nonzero(a)
    b_feat = np.count_nonzero(b)
    c = 0
    for idx, _ in enumerate(a):
        if a[idx] == b[idx] and a[idx] != 0:
            c += 1
    b_d = (2 * c) / (a_feat + b_feat)
    return b_d


def cosine(a, b):
    """Compute cosine coefficient.

    Parameters
    ----------
    a : array_like
        molecule A's features.
    b : array_like
        molecules B's features.

    Returns
    -------
    coeff : int
        cosine coefficient for molecule A and B.
    """
    coeff = (sum(a * b)) / (((sum(a**2)) + (sum(b**2))) ** 0.5)
    return coeff


def dice(a, b):
    """Compute dice coefficient.

    Parameters
    ----------
    a : array_like
        molecule A's features.
    b : array_like
        molecules B's features.

    Returns
    -------
    coeff : int
        dice coefficient for molecule A and B.
    """
    coeff = (2 * (sum(a * b))) / ((sum(a**2)) + (sum(b**2)))
    return coeff



# -*- coding: utf-8 -*-
#
# The Selector is a Python library of algorithms for selecting diverse
# subsets of data for machine-learning.
#
# Copyright (C) 2022-2024 The QC-Devs Community
#
# This file is part of Selector.
#
# Selector is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Selector is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Test Partition-Based Selection Methods."""

import numpy as np
import pytest
from numpy.testing import assert_equal, assert_raises

from selector.methods.partition import GridPartition, Medoid
from selector.methods.tests.common import generate_synthetic_data


@pytest.mark.parametrize("numb_pts", [4])
@pytest.mark.parametrize("method", ["equifrequent", "equisized"])
def test_grid_partitioning_independent_on_simple_example(numb_pts, method):
    r"""Test grid partitioning on an example each bin has only one point."""
    # Construct feature array where each molecule is known to be in which bin
    # The grid is a uniform grid from 0 to 10 in X-axis and 0 to 11 on y-axis.
    x = np.linspace(0, 3, numb_pts)
    y = np.linspace(0, 3, numb_pts)
    X, Y = np.meshgrid(x, y)
    grid = np.array([1.0, 0.0]) + np.vstack([X.ravel(), Y.ravel()]).T
    # Make one bin have an extra point
    grid = np.vstack((grid, np.array([1.1, 0.0])))

    # Here the number of cells should be equal to the number of points in each dimension
    #  excluding the extra point, so that the answer is unique/known.
    collector = GridPartition(nbins_axis=4, bin_method=f"{method}_independent")
    # Sort the points so that they're comparable to the expected answer.
    selected_ids = np.sort(collector.select(grid, size=len(grid) - 1))
    expected = np.arange(len(grid) - 1)
    assert_equal(selected_ids, expected)


def test_grid_partitioning_equisized_dependent_on_simple_example():
    r"""Test equisized_dependent grid partitioning on example that is different from independent."""
    # Construct feature array where each molecule is known to be in which bin
    grid = np.array(
        [
            [0.0, 0.0],  # Corresponds to bin (0, 0)
            [0.0, 4.0],  # Corresponds to bin (0, 3)
            [1.0, 1.0],  # Corresponds to bin (1, 0)
            [1.0, 0.9],  # Corresponds to bin (1, 0)
            [1.0, 2.0],  # Corresponds to bin (1, 3)
            [2.0, 0.0],  # Corresponds to bin (2, 0)
            [2.0, 4.0],  # Corresponds to bin (2, 3)
            [3.0, 0.0],  # Corresponds to bin (3, 0)
            [3.0, 4.0],  # Corresponds to bin (3, 3)
            [3.0, 3.9],  # Corresponds to bin (3, 3)
        ]
    )

    # The number of bins makes it so that it approximately be a single point in each bin
    collector = GridPartition(nbins_axis=4, bin_method="equisized_dependent")
    # Two bins have an extra point in them and so has more diversity than other bins
    #   then the two expected molecules should be in those bins.
    selected_ids = collector.select(grid, size=2, labels=None)
    right_molecules = True
    if not (2 in selected_ids or 3 in selected_ids):
        right_molecules = False
    if not (8 in selected_ids or 9 in selected_ids):
        right_molecules = False
    assert right_molecules, "The correct points were selected"


@pytest.mark.parametrize("numb_pts", [4])
def test_grid_partitioning_equifrequent_dependent_on_simple_example(numb_pts):
    r"""Test equifrequent dependent grid partitioning on an example where each bin has only one point."""
    # Construct feature array where each molecule is known to be in which bin
    # The grid is a uniform grid from 0 to 10 in X-axis and 0 to 11 on y-axis.
    x = np.linspace(0, 3, numb_pts)
    y = np.linspace(0, 3, numb_pts)
    X, Y = np.meshgrid(x, y)
    grid = np.array([1.0, 0.0]) + np.vstack([X.ravel(), Y.ravel()]).T
    # Make one bin have an extra point
    grid = np.vstack((grid, np.array([1.1, 0.0])))

    # Here the number of cells should be equal to the number of points in each dimension
    #  excluding the extra point, so that the answer is unique/known.
    collector = GridPartition(nbins_axis=numb_pts, bin_method="equifrequent_dependent")
    # Sort the points so that they're comparable to the expected answer.
    selected_ids = np.sort(collector.select(grid, size=len(grid) - 1))
    expected = np.arange(len(grid) - 1)
    assert_equal(selected_ids, expected)


@pytest.mark.parametrize("numb_pts", [10, 20, 30])
@pytest.mark.parametrize("method", ["equifrequent", "equisized"])
def test_bins_from_both_methods_dependent_same_as_independent_on_uniform_grid(numb_pts, method):
    r"""Test bins is the same between the two equisized methods on uniform grid in three-dimensions."""
    x = np.linspace(0, 10, numb_pts)
    y = np.linspace(0, 11, numb_pts)
    X = np.meshgrid(x, y, y)
    grid = np.vstack(list(map(np.ravel, X))).T
    grid = np.array([1.0, 0.0, 0.0]) + grid

    # Here the number of cells should be equal to the number of points in each dimension
    #  excluding the extra point, so that the answer is unique/known.
    collector_indept = GridPartition(nbins_axis=numb_pts, bin_method=f"{method}_independent")
    collector_depend = GridPartition(nbins_axis=numb_pts, bin_method=f"{method}_dependent")

    # Get the bins from the method
    bins_indept = collector_indept.get_bins_from_method(grid)
    bins_dept = collector_depend.get_bins_from_method(grid)

    # Test the bins are the same
    for key in bins_indept.keys():
        assert_equal(bins_dept[key], bins_indept[key])


def test_raises_grid_partitioning():
    r"""Test raises error for grid partitioning."""
    grid = np.random.uniform(0.0, 1.0, size=(10, 3))

    assert_raises(TypeError, GridPartition, 5.0)  # Test number of axis should be integer
    assert_raises(TypeError, GridPartition, 5, 5.0)  # Test grid method should be string
    assert_raises(TypeError, GridPartition, 5, "string", [])  # Test random seed should be integer

    # Test the collector grid method is not the correct string
    collector = GridPartition(nbins_axis=5, bin_method="string")
    assert_raises(ValueError, collector.select_from_cluster, grid, 5)

    collector = GridPartition(nbins_axis=5)
    assert_raises(TypeError, collector.select_from_cluster, [5.0], 5)  # Test X is numpy array
    assert_raises(
        TypeError, collector.select_from_cluster, grid, 5.0
    )  # Test number selected should be int
    assert_raises(TypeError, collector.select_from_cluster, grid, 5, [5.0])


def test_medoid():
    """Testing Medoid class."""
    coords, _, _ = generate_synthetic_data(
        n_samples=100,
        n_features=2,
        n_clusters=1,
        pairwise_dist=True,
        metric="euclidean",
        random_state=42,
    )

    coords_cluster, class_labels_cluster, _ = generate_synthetic_data(
        n_samples=100,
        n_features=2,
        n_clusters=3,
        pairwise_dist=True,
        metric="euclidean",
        random_state=42,
    )
    collector = Medoid()
    selected_ids = collector.select(coords_cluster, size=12, labels=class_labels_cluster)
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [2, 73, 94, 86, 1, 50, 93, 78, 0, 54, 33, 72])

    collector = Medoid()
    selected_ids = collector.select(coords, size=12)
    # make sure all the selected indices are the same with expectation
    assert_equal(selected_ids, [0, 95, 57, 41, 25, 9, 8, 6, 66, 1, 42, 82])

    # test the case where KD-Tree query return is an integer
    features = np.array([[1.5, 2.8], [2.3, 3.8], [1.5, 2.8], [4.0, 5.9]])
    selector = Medoid()
    selected_ids = selector.select(features, size=2)
    assert_equal(selected_ids, [0, 3])
